/*
 * Read and write "certdb" files, which are just text files containing
 * enaml lines.
 *
 * Copyright (C) 2001  Robey Pointer
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <glib.h>

#include "log.h"
#include "enaml.h"
#include "certdb.h"

/* format of cert lines:
 * server:{ host:"..." cert:"..." }
 */

#define HEADER "# say2 cert db\n# you probably don't want to edit this by hand.\n\n"

#define MAX_LINE	8192


Say2Cert *
say2_cert_new (void)
{
        Say2Cert *cert;

        cert = g_new0 (Say2Cert, 1);
        return cert;
}

Say2Cert *
say2_cert_new_host (const char *hostname, int length, const void *cert_data)
{
        Say2Cert *cert;

        cert = say2_cert_new ();
        cert->hostname = g_strdup (hostname);
        g_strdown (cert->hostname);
        cert->length = length;
        cert->cert = g_malloc (length);
        memcpy (cert->cert, cert_data, length);

        return cert;
}

void
say2_cert_add_alias (Say2Cert *cert, const char *hostname)
{
        GList *iter;
        char *hostname_copy;

        for (iter = g_list_first (cert->alias); iter != NULL; iter = g_list_next (iter)) {
                if (strcmp ((char *)iter->data, hostname) == 0) {
                        return;
                }
        }
        hostname_copy = g_strdup (hostname);
        g_strdown (hostname_copy);
        cert->alias = g_list_prepend (cert->alias, hostname_copy);
}

void
say2_cert_free (Say2Cert *cert)
{
        GList *iter;

        g_free (cert->hostname);
        g_free (cert->cert);
        for (iter = g_list_first (cert->alias); iter != NULL; iter = g_list_next (iter)) {
                g_free ((char *)iter->data);
        }
        g_list_free (cert->alias);

        cert->hostname = NULL;
        cert->alias = NULL;
        cert->length = 0;
        cert->cert = NULL;
}

gboolean
say2_cert_matches_host (Say2Cert *cert, const char *hostname)
{
        char *hostname_lower;
        gboolean ret = FALSE;
        GList *iter;

        hostname_lower = g_strdup (hostname);
        g_strdown (hostname_lower);
        if (strcmp (hostname_lower, cert->hostname) == 0) {
                ret = TRUE;
        } else {
                for (iter = g_list_first (cert->alias); iter != NULL; iter = g_list_next (iter)) {
                        if (strcmp (hostname_lower, (char *)iter->data) == 0) {
                                ret = TRUE;
                        }
                }
        }

        g_free (hostname_lower);
        return ret;
}
        
GHashTable *
say2_cert_table_new (void)
{
        return g_hash_table_new (g_str_hash, g_str_equal);
}

static gboolean
cert_remove_callback (const char *key, Say2Cert *value, void *userdata)
{
        value->refcount--;
        if (value->refcount == 0) {
                say2_cert_free (value);
        }
        return TRUE;
}

void
say2_cert_table_free (GHashTable *table)
{
        g_hash_table_foreach_remove (table, (GHRFunc)cert_remove_callback, NULL);
        g_hash_table_destroy (table);
}

static void
say2_cert_table_add_host (GHashTable *table, const char *hostname, Say2Cert *cert)
{
        Say2Cert *oldcert;

        /* previous cert already? */
        oldcert = (Say2Cert *) g_hash_table_lookup (table, hostname);
        if (oldcert != NULL) {
                g_warning ("duplicate cert for host '%s'", hostname);
                g_hash_table_remove (table, hostname);
                oldcert->refcount--;
                if (oldcert->refcount == 0) {
                        say2_cert_free (oldcert);
                }
        }

        cert->refcount++;
        g_hash_table_insert (table, (char *)hostname, cert);
}

void
say2_cert_table_add (GHashTable *table, Say2Cert *cert)
{
        GList *iter;

        /* hostnames in the cert are already all lowercase */
        say2_cert_table_add_host (table, cert->hostname, cert);
        for (iter = g_list_first (cert->alias); iter != NULL; iter = g_list_next (iter)) {
                say2_cert_table_add_host (table, (char *)iter->data, cert);
        }
}

Say2Cert *
say2_cert_table_lookup (GHashTable *table, const char *hostname)
{
        return (Say2Cert *) g_hash_table_lookup (table, hostname);
}

static void
table_foreach (char *hostname, Say2Cert *cert, GList **list)
{
        *list = g_list_prepend (*list, hostname);
}

GList *
say2_cert_table_get_hostnames (GHashTable *table)
{
        GList *list = NULL;

        g_hash_table_foreach (table, (GHFunc)table_foreach, &list);
        return list;
}

/* read a certdb file and return a GHashTable<char *host, Say2Cert *cert> */
gboolean
say2_cert_table_load (GHashTable *table, const char *filename)
{
        FILE *fp;
        char buffer[MAX_LINE];
        char *line;
        const char *hostname;
        int cert_len;
        const char *cert_data;
        Say2Cert *cert;
        EnamlBlock *block, *sub;
        EnamlList *alias_list;
        int i;

        fp = fopen (filename, "r");
        if (fp == NULL) {
                return FALSE;
        }

        while (! feof (fp)) {
                fgets (buffer, MAX_LINE, fp);
                if (feof (fp)) {
                        break;
                }
                buffer[MAX_LINE-1] = 0;

                line = buffer;
                while ((*line == ' ') || (*line == '\t')) {
                        line++;
                }
                if ((line[0] == '\0') || (line[0] == '#') || (line[0] == ';') ||
                    (line[0] == '\n') || (line[0] == '\r')) {
                        continue;
                }

                block = enaml_parse (line);
                if (block == NULL) {
                        g_warning ("invalid certdb line: %s", line);
                        continue;
                }

                sub = enaml_get_block (block, "host-cert");
                if (sub == NULL) {
                        g_warning ("unsupported certdb key: %s", line);
                        printf ("out: '%s'\n", enaml_compose (block, FALSE));
                        enaml_free (block);
                        continue;
                }

                hostname = enaml_get_string (sub, "host", NULL);
                cert_len = enaml_get_blob (sub, "cert", &cert_data);
                if ((hostname == NULL) || (cert_len == 0)) {
                        g_warning ("incomplete cert: %s", line);
                        enaml_free (block);
                        continue;
                }

                alias_list = enaml_get_list (sub, "alias");

#ifdef WHAT_THE_FUCK
                /* previous cert already? */
                cert = (Say2Cert *) g_hash_table_lookup (table, hostname_lower);
                if (cert != NULL) {
                        g_warning ("duplicate cert for host '%s'", hostname_lower);
                        g_hash_table_remove (table, hostname_lower);
                        say2_cert_free (cert);
                }
#endif

                cert = say2_cert_new_host (hostname, cert_len, cert_data);
                if (alias_list != NULL) {
                        for (i = 0; i < enaml_list_length (alias_list); i++) {
                                say2_cert_add_alias (cert, enaml_list_get_string (alias_list, i, "??"));
                        }
                }
                say2_cert_table_add (table, cert);

                /* hostname_lower belongs to cert now */
                enaml_free (block);
        }

        fclose (fp);
        return TRUE;
}

static void
cert_write_callback (const char *hostname, Say2Cert *cert, int *fd)
{
        EnamlBlock *block;
        EnamlList *alias_list;
        char *line;
        GList *iter;

        if (strcmp (hostname, cert->hostname) != 0) {
                /* only write each entry once please! */
                return;
        }

        alias_list = NULL;
        if (cert->alias != NULL) {
                alias_list = enaml_new_list ();
                for (iter = g_list_first (cert->alias); iter != NULL; iter = g_list_next (iter)) {
                        enaml_list_add_string (alias_list, (char *)iter->data, 0);
                }
        }
        block = enaml_buildf ("host-cert", "host:%s cert:%x alias:%L", cert->hostname, cert->cert, cert->length,
                              alias_list);
        if (block == NULL) {
                line = g_strdup_printf ("# ERROR building %s\n", cert->hostname);
                if (*fd > 0) {
                        if (write (*fd, line, strlen (line)) < strlen (line)) {
                                *fd = -1;
                        }
                }
                g_free (line);
                return;
        }
        line = enaml_compose (block, FALSE);
        if (*fd > 0) {
                if (write (*fd, line, strlen (line)) < strlen (line)) {
                        *fd = -1;
                }
        }
        if (*fd > 0) {
                if (write (*fd, "\n", 1) < 1) {
                        *fd = -1;
                }
        }
        g_free (line);
        enaml_free (block);
}

gboolean
say2_cert_table_save_fd (GHashTable *table, int fd)
{
        int sub_fd;

        ftruncate (fd, 0);
        lseek (fd, 0, SEEK_SET);
        write (fd, HEADER, strlen (HEADER));
        sub_fd = fd;
        g_hash_table_foreach (table, (GHFunc)cert_write_callback, &sub_fd);
        if (sub_fd == fd) {
                write (fd, "\n", 1);
                return TRUE;
        } else {
                return FALSE;
        }
}

gboolean
say2_cert_table_save (GHashTable *table, const char *filename)
{
        int fd;
        gboolean ret;

        fd = open (filename, O_CREAT|O_WRONLY, 0600);
        if (fd < 0) {
                return FALSE;
        }

        ret = say2_cert_table_save_fd (table, fd);
        close (fd);
        return ret;
}
