/*
 * This is a sort of generalization of the "command table" setup that many
 * servers have.  You submit a table of command names and function pointers,
 * they're hashed into a real table, and then you call back with a command
 * and up to 3 pointers to pass to the command functions.
 *
 * This particular implementation assumes that the commands are case-neutral.
 * It's very simplistic because it's mostly based on glib's hashtables.
 *
 *
 * Copyright (C) 2001  Robey Pointer <robey@lag.net>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "dispatch.h"

typedef void (*DispatchFunction)(void *arg1, void *arg2, void *arg3);

/* dump a list of dispatch items into a hashtable */
void *
dispatch_create (DispatchItem *list)
{
        GHashTable *table;
        char *command;
        int i;

        table = g_hash_table_new (g_str_hash, g_str_equal);
        for (i = 0; list[i].command != NULL; i++) {
                command = g_strdup (list[i].command);
                g_strdown (command);
                g_hash_table_insert (table, command, list[i].function);
        }

        return table;
}

/*
 * Find a command in the table, and call the matching function.  The three
 * args (arg1, arg2, arg3) are passed as the three parameters, and it's assumed
 * that the function has no return value.  (You can use one of the 3 args as a
 * pass-by-reference return value if you choose.)
 *
 * Returns TRUE if any command matched, FALSE if not.
 */
gboolean
dispatch_go (void *xtable, const char *command, void *arg1, void *arg2, void *arg3)
{
        GHashTable *table;
        char *command_lower;
        void *ptr;
        DispatchFunction func;

        table = (GHashTable *)xtable;
        command_lower = g_strdup (command);
        g_strdown (command_lower);

        ptr = g_hash_table_lookup (table, command_lower);
        g_free (command_lower);

        if (ptr == NULL) {
                return FALSE;
        }

        func = (DispatchFunction)ptr;
        (*func)(arg1, arg2, arg3);
        return TRUE;
}

static gboolean
dispatch_foreach_remove (char *key, void *value)
{
        g_free (key);
        return TRUE;	/* remove this key */
}

void
dispatch_free (void *xtable)
{
        GHashTable *table;

        table = (GHashTable *)xtable;
        g_hash_table_foreach_remove (table, (GHRFunc)dispatch_foreach_remove, NULL);
        g_hash_table_destroy (table);
}
