/* $Id: dnscache.c,v 1.6 2001/05/27 20:36:50 robey Exp $
 *
 * This is a caching DNS system. When a host name is needed we look it up here
 * and see if there is already an answer for it. The domains are placed in a
 * hashed linked list. If the name is not here, then we need to look it up and
 * add it to the system. This really speeds up the connection to servers since
 * the DNS name does not need to be looked up each time. It's kind of cool. :)
 *
 * Copyright (C) 1999  Robert James Kaes (rjkaes@flarenet.com)
 * Copyright (C) 2000  Chris Lightfoot (chris@ex-parrot.com)
 * Copyright (C) 2000  Robey Pointer <robey@lag.net>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <assert.h>
#include <ctype.h>
#include <sys/types.h>
#include <glib.h>

#include "dnscache.h"

#ifndef INADDR_NONE
/* solaris bug */
#define INADDR_NONE (unsigned long)-1
#endif


#define HASH_BOX 31
#define DNSEXPIRE (5 * 60)

struct dnscache_s {
	struct in_addr ipaddr;
	char *domain;
	time_t expire;
	struct dnscache_s *next;
        struct dnscache_s *next_ip;
};

struct dnscache_s *cache[HASH_BOX];
struct dnscache_s *cache_reverse[HASH_BOX];

static unsigned int
hash (const unsigned char *keystr, unsigned int box)
{
	unsigned long hashc = 0;
	unsigned char ch;

	g_assert (keystr != NULL);
	g_assert (box > 0);

	while ((ch = *keystr++))
		hashc += tolower(ch);

	return hashc % box;
}

static unsigned int
hash_ip (const struct in_addr ipaddr, unsigned int box)
{
        return (unsigned long)(ipaddr.s_addr) % box;
}

static void
expired (struct dnscache_s *entry)
{
        unsigned int box = hash (entry->domain, HASH_BOX);
        unsigned int box_ip = hash_ip (entry->ipaddr, HASH_BOX);
        struct dnscache_s **ptr;

        /* remove from domain hash */
        for (ptr = &cache[box]; (*ptr) && (*ptr != entry); ptr = &(*ptr)->next)
                ;
        *ptr = (*ptr)->next;

        /* remove from ip hash */
        for (ptr = &cache_reverse[box_ip]; (*ptr) && (*ptr != entry);
             ptr = &(*ptr)->next_ip)
                ;
        *ptr = (*ptr)->next_ip;

        g_free (entry->domain);
        g_free (entry);
}

static int
lookup (struct in_addr *addr, const char *domain)
{
	unsigned int box = hash (domain, HASH_BOX);
	struct dnscache_s *ptr = cache[box];

	g_assert (domain);

	while (ptr && (g_strcasecmp (ptr->domain, domain) != 0)) {
		ptr = ptr->next;
	}

	if (ptr && (g_strcasecmp (ptr->domain, domain) == 0)) {
		/* Woohoo... found it. Make sure it hasn't expired */
		if (difftime(time(NULL), ptr->expire) > DNSEXPIRE) {
			/* Oops... expired */
                        expired (ptr);
			return -1;
		}

		/* chris - added this so that the routine can be used to just
		 * look stuff up.
		*/
		if (addr) {
			*addr = ptr->ipaddr;
                }
		return 0;
	}

	return -1;
}

static int
lookup_reverse (struct in_addr *addr, const char **domain)
{
	unsigned int box = hash_ip(*addr, HASH_BOX);
	struct dnscache_s *ptr = cache_reverse[box];

	while (ptr && (memcmp (&ptr->ipaddr, addr, sizeof (struct in_addr)) == 0)) {
		ptr = ptr->next_ip;
	}

        if (ptr && (memcmp (&ptr->ipaddr, addr, sizeof (struct in_addr)) == 0)) {
		/* Woohoo... found it. Make sure it hasn't expired */
		if (difftime(time(NULL), ptr->expire) > DNSEXPIRE) {
			/* Oops... expired */
                        expired (ptr);
			return -1;
		}
                if (domain) {
                        *domain = ptr->domain;
                }
		return 0;
	}

	return -1;
}

static int
insert (struct in_addr *addr, const char *domain)
{
	unsigned int box = hash (domain, HASH_BOX);
        unsigned int box_reverse = hash_ip (*addr, HASH_BOX);
	struct dnscache_s *newptr;

	g_assert (addr != NULL);
	g_assert (domain != NULL);

	newptr = g_new0 (struct dnscache_s, 1);

	if (!(newptr->domain = g_strdup (domain))) {
		g_free (newptr);
		return -1;
	}

	newptr->ipaddr = *addr;
	newptr->expire = time(NULL);
        newptr->next = cache[box];
        newptr->next_ip = cache_reverse[box_reverse];
        cache[box] = newptr;
        cache_reverse[box_reverse] = newptr;

	return 0;
}

int
dnscache_lookup (struct in_addr *addr, const char *domain)
{
	struct hostent *resolv;

	g_return_val_if_fail (addr != NULL, -1);
	g_return_val_if_fail (domain != NULL, -1);

#ifdef HAVE_INET_ATON
        if (inet_aton(domain, (struct in_addr *) addr) != 0)
                return 0;
#else
        /* try inet_addr instead */
        if (inet_addr(domain) != INADDR_NONE) {
                addr->s_addr = inet_addr(domain);
                return 0;
        }
#endif

	/* Well, we're not dotted-decimal so we need to look it up */
	if (lookup(addr, domain) == 0)
		return 0;

	/* Okay, so not in the list... need to actually look it up. */
	if (!(resolv = gethostbyname(domain)))
		return -1;

	memcpy(addr, resolv->h_addr_list[0], resolv->h_length);
	insert(addr, domain);

	return 0;
}

int
dnscache_lookup_reverse (struct in_addr *addr, const char **domain)
{
        struct hostent *resolv;
        struct in_addr addr2;

        g_return_val_if_fail (addr != NULL, -1);
        g_return_val_if_fail (domain != NULL, -1);

        if (lookup_reverse (addr, domain) == 0) {
                return 0;
        }

        if (!(resolv = gethostbyaddr ((char *)addr, sizeof (struct in_addr),
                                      AF_INET))) {
                *domain = inet_ntoa (*addr);
                return 0;
        }

        /* reverse names are often faked -- double-check */
        if ((dnscache_lookup (&addr2, resolv->h_name) == 0) &&
            (memcmp (addr, &addr2, sizeof (struct in_addr)) == 0)) {
                *domain = resolv->h_name;
                return 0;
        }

        /* fake reverse lookup */
        *domain = inet_ntoa (*addr);
        return 0;
}

void
dnscache_cleanup (void)
{
	unsigned int c;
	struct dnscache_s *ptr, *tmp;

	for (c = 0; c < HASH_BOX; c++) {
		ptr = cache[c];

		while (ptr) {
			tmp = ptr->next;

			if (difftime(time(NULL), ptr->expire) > DNSEXPIRE) {
				expired (ptr);
                        }

			ptr = tmp;
		}
	}
}
