/*
 * Composes enaml blocks into strings.
 *
 * Copyright (C) 2000  Robey Pointer <robey@lag.net>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include "enaml.h"

/* only quote these when debugging */
#define NEEDS_QUOTE_DEBUG(c) \
	(NEEDS_QUOTE_NORMAL(c) || ((c >= 0x80) && (c <= 0xA0)))
/* add 0xFF for telnet clients */
#define NEEDS_QUOTE_NORMAL(c) \
	((c < 0x20) || (c == 0x7F) || (c == 0x22) || (c == 0x25) || (c == 0x27) || (c == 0xFF))

#define NEEDS_QUOTE(c)  NEEDS_QUOTE_NORMAL(c)

static int value_size (EnamlValue *value);
static char *write_value (char *out, EnamlValue *value);

static void
block_count_func (const char *key, EnamlValue *value, int *size)
{
        int vsize;

        *size += strlen (key);
        vsize = value_size (value) + 1;		/* " " */
        if (vsize) {
                *size += vsize + 1;		/* ":" */
        }
}

static int
block_size (const EnamlBlock *block)
{
        int size = 0;

        if (block == NULL) {
                return 0;
        }
        g_hash_table_foreach ((GHashTable *)block, (GHFunc)block_count_func, (void *)&size);
        return size;
}

static int
list_size (EnamlList *list)
{
        int size = 0;
        GList *iter;

        for (iter = g_list_first (*list); iter; iter = g_list_next (iter)) {
                size += value_size ((EnamlValue *)(iter->data)) + 1;	/* " " */
        }
        return size;
}

/* count the chars in an EnamlValue's data block that will need to be quoted, and
 * come up with an upper bound on the space needed to store the data (NOT including
 * the key) in a string.
 */
static int
value_size (EnamlValue *value)
{
        int i;

        if (value->block) {
                return block_size (value->block) + 3;		/* "{ " + "}" */
        }
        if (value->list) {
                return list_size (value->list) + 3;		/* "[ " + "]" */
        }
        if (value->string) {
                value->flags |= ENAML_DIGITS;
                value->quotes = 0;

                for (i = 0; i < value->length; i++) {
                        unsigned char c = (unsigned char)(value->string[i]);

                        if (NEEDS_QUOTE (c)) {
                                value->quotes++;
                        }
                        if ((value->flags & ENAML_DIGITS) && ((c < '0') || (c > '9'))) {
                                value->flags &= ~ENAML_DIGITS;
                        }
                }
                if (value->length == 0) {
                        value->flags &= ~ENAML_DIGITS;
                }

                /* if over 50% have to be quoted, elect to do a %-hex quote */
                if (value->quotes > value->length/2) {
                        return (value->length * 2) + 1;
                } else {
                        return (value->quotes * 3) + (value->length - value->quotes) +
                                ((value->flags & ENAML_DIGITS) ? 0 : 2);
                }
        }
        return 0;
}


static void
block_write_func (const char *key, EnamlValue *value, char **pout)
{
        char *out = (*pout);

        strcpy (out, key);
        out += strlen (key);
        if (value->block || value->list || value->string) {
                *out++ = ':';
                out = write_value (out, value);
        }
        *out++ = ' ';
        *pout = out;
}

static char *
write_block (char *out, const EnamlBlock *block)
{
        if (block != NULL) {
                g_hash_table_foreach ((GHashTable *)block, (GHFunc)block_write_func, &out);
        }
        return out;
}

static char *
write_list (char *out, EnamlList *list)
{
        GList *iter;

        for (iter = g_list_first (*list); iter; iter = g_list_next (iter)) {
                out = write_value (out, (EnamlValue *)(iter->data));
                *out++ = ' ';
        }
        return out;
}

static char *
write_value (char *out, EnamlValue *value)
{
        static char *hex = "0123456789ABCDEF";
        int i;

        if (value->block) {
                *out++ = '{';
                *out++ = ' ';
                out = write_block (out, value->block);
                *out++ = '}';
                return out;
        }
        if (value->list) {
                *out++ = '[';
                *out++ = ' ';
                out = write_list (out, value->list);
                *out++ = ']';
                return out;
        }
        if (value->string) {
                if (value->quotes > value->length/2) {
                        /* quote the whole thing */
                        *out++ = '%';
                        for (i = 0; i < value->length; i++) {
                                unsigned char c = (unsigned char)(value->string[i]);
                                *out++ = hex[c/16];
                                *out++ = hex[c%16];
                        }
                } else {
                        /* no need to add quotes if it's all digits */
                        if (! (value->flags & ENAML_DIGITS)) {
                                *out++ = '"';
                        }
                        for (i = 0; i < value->length; i++) {
                                unsigned char c = (unsigned char)(value->string[i]);

                                if (NEEDS_QUOTE (c)) {
                                        *out++ = '%';
                                        *out++ = hex[c/16];
                                        *out++ = hex[c%16];
                                } else {
                                        *out++ = value->string[i];
                                }
                        }
                        if (! (value->flags & ENAML_DIGITS)) {
                                *out++ = '"';
                        }
                }
                return out;
        }

        /* do nothing then. */
        return out;
}


/* allocates a string that's long enough to hold this block, and writes an
 * ENAML protocol message into it -- you need to free the string when done
 */
char *
enaml_compose (const EnamlBlock *block, int add_cr)
{
        int size = block_size (block);
        char *out;
        char *p;

	if (add_cr) {
		size += 2;
	}
	out = (char *) g_malloc (size + 1);
        p = write_block (out, block);
        if (add_cr) {
                *p++ = 13;
                *p++ = 10;
        }
        *p = 0;

        g_assert (strlen (out) <= size);

        return out;
}

/* frees a string created with enaml_compose() */
void
enaml_compose_free (char *s)
{
        g_free (s);
}
