/*
 * Parsing function of enaml.  This used to be in lex, but it seemed to
 * be overkill for such a simple parser.
 *
 * Copyright (C) 2000  Robey Pointer <robey@lag.net>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 *
 * Written as a whole on 6 September 2000: Be wary of late-night mochas.
 *
 * Notes: I tried to replace the tokenizer function with lex, but lex was
 * slower.  That surprised the hell out of me since this parser is half as
 * fast as the old lex-generated one.  Getting rid of the recursion didn't
 * help either.  I'm left with the inevitable conclusion that the glib
 * list/hash functions are slower than my old ones.  Oh well.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include "enaml.h"

/* max nesting allowed for blocks and lists */
#define MAX_DEPTH	32


static char token_table[256];
static int token_table_initialized = 0;
#define IS_WHITESPACE	1
#define IS_DIGIT	2
#define IS_KEY		4
#define IS_HEX		8
#define WHITESPACE(x)	(token_table[(int)x] & IS_WHITESPACE)
#define DIGIT(x)	(token_table[(int)x] & IS_DIGIT)
#define KEY(x)		(token_table[(int)x] & IS_KEY)
#define HEX(x)		(token_table[(int)x] & IS_HEX)

typedef enum {
        TOKEN_UNKNOWN = -1,
        TOKEN_EOF = 0,
        TOKEN_OPEN_BRACE,
        TOKEN_CLOSE_BRACE,
	TOKEN_OPEN_BRACKET,
        TOKEN_CLOSE_BRACKET,
        TOKEN_COLON,
        TOKEN_NUMBER,
        TOKEN_STRING,
        TOKEN_HEX,
        TOKEN_KEY
} Token;

#define DONE(_token) do { \
	*string = p+1; \
	return _token; \
} while (0)


#define unhex(x) ( (((x) >= 'A') && ((x) <= 'F')) ? ((x)-'A'+10) : \
                   ((((x) >= 'a') && ((x) <= 'f')) ? ((x)-'a'+10) : ((x)-'0')))
static int
unhtmlify (char *in)
{
        char *p = in;
        int len = strlen (in);
        int i;

        for (i = 0; i < len; i++) {
                if ((in[i] == '%') && (i+2 < len)) {
                        *p++ = unhex (in[i+1]) * 16 + unhex (in[i+2]);
                        i += 2;
                } else {
                        *p++ = in[i];
                }
        }
        *p = 0;
        return p - in;
}

static int
unhexify (char *in)
{
        char *p = in;
        int len = strlen (in);
        int i;

        for (i = 0; i < len; i += 2) {
                *p++ = unhex (in[i]) * 16 + unhex (in[i+1]);
        }
        *p = 0;
        return len/2;
}

static Token
next_token (const char **string, char **value)
{
        const char *p = *string;
        const char *q;

        *value = NULL;
        if (! p) {
                return TOKEN_EOF;
        }
        while (*p && WHITESPACE(*p)) {
                p++;
        }
        if (! *p) {
                return TOKEN_EOF;
        }

        switch (*p) {
        case '{':
                DONE (TOKEN_OPEN_BRACE);
        case '}':
                DONE (TOKEN_CLOSE_BRACE);
        case '[':
                DONE (TOKEN_OPEN_BRACKET);
        case ']':
                DONE (TOKEN_CLOSE_BRACKET);
        case ':':
                DONE (TOKEN_COLON);
        }

        if (*p == '%') {
                p++;
                for (q = p; *q && HEX (*q); q++)
                        ;
                *value = g_strndup (p, q-p);
                *string = q;
                return TOKEN_HEX;
        }

        if ((*p == '"') || (*p == '\'')) {
                for (q = p+1; *q && (*q != *p); q++)
                        ;
                p++;
                *value = g_strndup (p, q-p);
                *string = (*q ? q+1 : q);
                return TOKEN_STRING;
        }

        if (DIGIT (*p)) {
                for (q = p; *q && DIGIT (*q); q++)
                        ;
                *value = g_strndup (p, q-p);
                *string = q;
                return TOKEN_NUMBER;
        }

        if (KEY (*p)) {
                for (q = p; *q && KEY (*q); q++)
                        ;
                *value = g_strndup (p, q-p);
                *string = q;
                return TOKEN_KEY;
        }

        DONE (TOKEN_UNKNOWN);
}

static void
generate_table (void)
{
        int i;

        for (i = 0; i < 256; i++) {
                token_table[i] = 0;
        }
        for (i = '0'; i <= '9'; i++) {
                token_table[i] |= (IS_DIGIT | IS_HEX);
        }
        token_table[' '] |= IS_WHITESPACE;
        token_table['\t'] |= IS_WHITESPACE;
        token_table[10] |= IS_WHITESPACE;
        token_table[13] |= IS_WHITESPACE;
        for (i = 'a'; i <= 'f'; i++) {
                token_table[i] |= (IS_KEY | IS_HEX);
        }
        for (i = 'g'; i <= 'z'; i++) {
                token_table[i] |= IS_KEY;
        }
        for (i = 'A'; i <= 'F'; i++) {
                token_table[i] |= (IS_KEY | IS_HEX);
        }
        for (i = 'G'; i <= 'Z'; i++) {
                token_table[i] |= IS_KEY;
        }
        token_table['-'] |= IS_KEY;
        token_table['_'] |= IS_KEY;

        token_table_initialized = 1;
}


typedef enum {
        STATE_BLOCK,
        STATE_LIST
} ParserState;

EnamlBlock *
enaml_parse (const char *string)
{
        ParserState state = STATE_BLOCK;
        GList *block_stack = NULL;
        GList *list_stack = NULL;
        Token token;
        EnamlBlock *block;
        EnamlList *list = NULL;
        char *key = NULL;
        char *value;
        int len;

        if (! token_table_initialized) {
                generate_table ();
        }

        block = enaml_new ();
        while (1) {
                token = next_token (&string, &value);
                switch (token) {
                case TOKEN_OPEN_BRACE:
                        if (! key || (state != STATE_BLOCK)) {
                                goto fail;
                        }
                        if (g_list_length (block_stack) >= MAX_DEPTH) {
                                goto fail;
                        }
                        block_stack = g_list_prepend (block_stack, block);
                        block = enaml_new ();
                        enaml_add_block ((EnamlBlock *) (block_stack->data), key, block, 
                                         ENAML_FREE);
                        /* back to get-key state */
                        g_free (key);
                        key = NULL;
                        break;
                case TOKEN_CLOSE_BRACE:
                        if (!block_stack || (state != STATE_BLOCK)) {
                                goto fail;
                        }
                        if (key) {
                                enaml_add_flag (block, key);
                                g_free (key);
                                key = NULL;
                        }
                        block = block_stack->data;
                        block_stack = g_list_remove (block_stack, block_stack->data);
                        break;
                case TOKEN_OPEN_BRACKET:
                        if (state == STATE_BLOCK) {
                                if (!key) {
                                        goto fail;
                                }
                                list = enaml_new_list ();
                                enaml_add_list (block, key, list, ENAML_FREE);
                                state = STATE_LIST;
                                g_free (key);
                                key = NULL;
                        } else {	/* STATE_LIST */
                                if (g_list_length (list_stack) >= MAX_DEPTH) {
                                        goto fail;
                                }
                                list_stack = g_list_prepend (list_stack, list);
                                list = enaml_new_list ();
                                enaml_list_add_list ((EnamlList *) (list_stack->data), list,
                                                     ENAML_FREE);
                        }
                        break;
                case TOKEN_CLOSE_BRACKET:
                        if (state != STATE_LIST) {
                                goto fail;
                        }
                        if (list_stack) {
                                list = list_stack->data;
                                list_stack = g_list_remove (list_stack, list_stack->data);
                        } else {
                                list = NULL;
                                state = STATE_BLOCK;
                        }
                        break;
                case TOKEN_COLON:
                        if ((state != STATE_BLOCK) || !key) {
                                goto fail;
                        }
                        break;
                case TOKEN_NUMBER:
                case TOKEN_STRING:
                case TOKEN_HEX:
                        if ((state == STATE_BLOCK) && !key) {
                                goto fail;
                        }
                        if (token == TOKEN_HEX) {
                                len = unhexify (value);
                        } else {
                                len = unhtmlify (value);
                        }
                        if (state == STATE_BLOCK) {
                                enaml_add_blob (block, key, value, len, ENAML_FREE);
                                g_free (key);
                                key = NULL;
                        } else {
                                enaml_list_add_blob (list, value, len, ENAML_FREE);
                        }
                        break;
                case TOKEN_KEY:
                        if (state == STATE_LIST) {
                                goto fail;
                        }
                        if (key) {
                                enaml_add_flag (block, key);
                                g_free (key);
                        }
                        key = value;
                        break;
                case TOKEN_UNKNOWN:
                        goto fail;
                case TOKEN_EOF:
                        goto pass;
                }
        }

pass:
        /* clean up any unfinished business */
        if (key) {
                enaml_add_flag (block, key);
                g_free (key);
        }
        /* implicitly close any remaining lists/blocks */
        g_list_free (list_stack);
        while (block_stack) {
                block = (EnamlBlock *) (block_stack->data);
                block_stack = g_list_remove (block_stack, block_stack->data);
        }
        return block;
        
fail:
        g_free (key);
        g_list_free (list_stack);
        while (block_stack) {
                block = (EnamlBlock *) (block_stack->data);
                block_stack = g_list_remove (block_stack, block_stack->data);
        }
        if (block) {
                enaml_free (block);
        }
        return NULL;
}



/* build an enaml block out of a format string.  example:
 *     block = enaml_buildf ("test", "name:%s coords:%2d present age:%d",
 *                           "Sam Lowry", 23, 9, 37);
 *     --> test:{ name:"Sam Lowry" coords:[23 9] present age:37 }
 * possible %-flags are:
 *     f: flag (default if %-flag is missing)
 *     F: flag w/arg (see below)
 *     d: long int
 *     s: string
 *     S: string that should be copied
 *     x: blob (buffer followed by length)
 *     X: blob that should be copied
 *     L: list
 *     B: block
 * a '!' after the '%' means to NOT set the ENAML_FREE flag for an L/B param.
 * a number before a d/s/S/L means to make a list out of that param.
 * a '*' before a d/s/S/L means to make a list out of that param, and that
 *     the count of list members is the next parameter.
 * for s/S/L/B, if the parameter is NULL, the key will be skipped (you can
 *     make keys optional this way).
 * for F, a gboolean parameter determines whether the key even appears at all
 *     (for TRUE, the flag appears; for FALSE, the flag is skipped)
 * if the key is %s, it will be pulled from the parameter list.
 */
EnamlBlock *
enaml_vbuildf (const char *tag, const char *fmt, va_list arg)
{
        EnamlBlock *out, *block, *myblock;
        EnamlList *list, *mylist;
        const char *p, *q, *key, *endkey, *string;
        char *dupkey;
        int count, do_free, type, i, num;

        out = enaml_new ();
        if (tag) {
                block = enaml_new ();
                enaml_add_block (out, tag, block, ENAML_FREE);
        } else {
                block = out;
        }

        for (key = fmt; key && *key; key = q) {
                count = 0;
                type = 'f';
                dupkey = NULL;
                do_free = 1;

                /* find end of this tag */
                q = strchr (key, ' ');
                if (q == NULL) {
                        q = key + strlen (key);
                }
                for (p = key; *p && (*p != ':') && (p < q); p++)
                        ;
                endkey = p;
                if (strncmp (key, "%s", 2) == 0) {
                        dupkey = g_strdup (va_arg (arg, char *));
                }
                if (*p == ':') {
                        p++;
                        /* parse p now, getting the flag type and count */
                        if (*p == '%') {
                                /* optional '%' to make it look like sprintf */
                                p++;
                        }
                        if (*p == '!') {
                                do_free = 0;
                                p++;
                        }
                        while ((*p >= '0') && (*p <= '9')) {
                                count = (count*10) + (*p - '0');
                                p++;
                        }
                        if (*p == '*') {
                                /* get count from params */
                                count = va_arg (arg, int);
                                p++;
                        }
                        switch (*p) {
                        case 'f':
                        case 'F':
                        case 'd':
                        case 's':
                        case 'S':
                        case 'x':
                        case 'X':
                        case 'L':
                        case 'B':
                                type = *p;
                                break;
                        case 'i':
                                type = 'd';
                                break;
                        }
                }

                if (count == 0) {
                        count = 1;
                }
                if (dupkey == NULL) {
                        dupkey = g_strndup (key, endkey - key);
                }
                if (count > 1) {
                        list = enaml_new_list ();
                } else {
                        list = NULL;
                }
                for (i = 0; i < count; i++) {
                        switch (type) {
                        case 'f':
                                enaml_add_flag (block, dupkey);
                                break;
                        case 'F':
                                if (va_arg (arg, gboolean)) {
                                        enaml_add_flag (block, dupkey);
                                }
                                break;
                        case 'd':
                                if (list) {
                                        enaml_list_add_long (list, va_arg (arg, long));
                                } else {
                                        enaml_add_long (block, dupkey, va_arg (arg, long));
                                }
                                break;
                        case 's':
                        case 'S':
                                string = va_arg (arg, const char *);
                                if (string != NULL) {
                                        if (list) {
                                                enaml_list_add_string (list, string,
                                                                       type == 's' ?
                                                                       0 : ENAML_COPY);
                                        } else {
                                                enaml_add_string (block, dupkey, string,
                                                                  type == 's' ?
                                                                  0 : ENAML_COPY);
                                        }
                                }
                                break;
                        case 'x':
                        case 'X':
                                string = va_arg (arg, const char *);
                                num = va_arg (arg, int);
                                if (string != NULL) {
                                        if (list) {
                                                enaml_list_add_blob (list, string, num,
                                                                     type == 'x' ? 0 : ENAML_COPY);
                                        } else {
                                                enaml_add_blob (block, dupkey, string, num,
                                                                type == 'x' ? 0 : ENAML_COPY);
                                        }
                                }
                                break;
                        case 'L':
                                mylist = va_arg (arg, EnamlList *);
                                if (mylist != NULL) {
                                        if (list) {
                                                enaml_list_add_list (list, mylist,
                                                                     do_free ? ENAML_FREE : 0);
                                        } else {
                                                enaml_add_list (block, dupkey, mylist,
                                                                do_free ? ENAML_FREE : 0);
                                        }
                                }
                                break;
                        case 'B':
                                myblock = va_arg (arg, EnamlBlock *);
                                if (myblock != NULL) {
                                        enaml_add_block (block, dupkey, myblock,
                                                         do_free ? ENAML_FREE : 0);
                                }
                                break;
                        }
                }
                if (count > 1) {
                        enaml_add_list (block, dupkey, list, ENAML_FREE);
                }
                g_free (dupkey);

                for ( ; *q && (*q == ' '); q++)
                        ;
        }

        if ((tag != NULL) && (enaml_block_size (block) == 0)) {
                /* top-level tag contains just an empty block: go ahead and replace with a flag */
                enaml_drop_key (out, tag);
                enaml_add_flag (out, tag);
        }

        return out;
}

EnamlBlock *
enaml_buildf (const char *tag, const char *fmt, ...)
{
        va_list arg;
        EnamlBlock *out;

        va_start (arg, fmt);
        out = enaml_vbuildf (tag, fmt, arg);
        va_end (arg);

        return out;
}
