/*
 * ENAML's Not A Markup Language
 *
 * Copyright (C) 2000  Robey Pointer <robey@lag.net>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 *
 * 1.0: 17apr2000, robey
 * 1.1: 18apr2000, robey
 *   * added support for lists of items [ .. ]
 * 1.1b: 19apr2000, robey
 *   * added support for value-less keys (no trailing ':')
 * 1.1c: 23apr2000, robey
 *   * allow enaml_get_*() functions to use compound keys to pull out
 *     nested members.  (for example, "login.userid" would pull out the
 *     "userid" field in the "login" block, if both existed.)
 *   * fix enaml_get_blob() to differentiate between a missing key, and
 *     a key with no value
 * 1.1d: 17may2000, robey
 *   * added the ability to iterate through the keys in a block
 *   * enaml_type() will return a flag indicating if a given key is a
 *     block, value, list, flag, or simply doesn't exist
 * 1.1e: 17aug2000, robey
 *   * removed enaml_done() which was just an exact duplicate of enaml_free()
 *   * enaml_put*() wasn't correctly freeing old pairs when overwriting,
 *     especially for lists and non-allocated values
 *   * results of enaml_parse() can now be changed (adding/removing pairs, etc)
 *     and fed into later enaml_compose()
 *   * cleaned up block operations into enaml_new_block()/enaml_put_block()/
 *     enaml_copy_block() (was a big mess)
 * 1.2: 6sep2000, robey
 *   * total rewrite of internal functions and structures, using glib and
 *     ditching lex
 *   * support for nested lists
 *
 *
 * TODO:
 *   * currently this is twice as slow as the old enaml parser :(
 *     (120us/parse vs 250us/parse)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include "enaml.h"


/* case-insensitive hash */
/* it also transposes digits and '-' onto the letters, but that's ok */
static guint
block_hash_fn (gconstpointer key)
{
        guint hash = 0;
        const char *p;

        for (p = (const char *)key; *p; p++) {
                hash = (hash << 3);
                hash += (*p & 0x1f);
        }
        return hash;
}

static gint
block_compare_fn (gconstpointer a, gconstpointer b)
{
        return (g_strcasecmp ((const char *)a, (const char *)b) == 0) ?
                TRUE : FALSE;
}

EnamlBlock *
enaml_new (void)
{
        return g_hash_table_new (block_hash_fn, block_compare_fn);
}

static void
free_value (EnamlValue *value)
{
        if (value->block) {
                if (value->flags & ENAML_FREE) {
                        enaml_free (value->block);
                }
                value->block = NULL;
        } else if (value->list) {
                if (value->flags & ENAML_FREE) {
                        enaml_free_list (value->list);
                }
                value->list = NULL;
        } else {
                if (value->flags & ENAML_FREE) {
                        g_free (value->string);
                }
                value->string = NULL;
        }
        value->flags = 0;
        value->quotes = 0;
        value->key[0] = 0;

        g_free (value);
}

EnamlList *
enaml_new_list (void)
{
        EnamlList *list;

        /* it's ok in glib to use NULL as the empty list */
        list = g_new0 (EnamlList, 1);
        return list;
}

void
enaml_free_list (EnamlList *list)
{
        g_list_foreach (*list, (GFunc)free_value, NULL);
}

static gboolean
free_block_item (char *key, EnamlValue *value, void *unused)
{
        free_value (value);
        return TRUE;
}

void
enaml_free (EnamlBlock *block)
{
        if (block == NULL) {
                /* already done, ma! */
                return;
        }
        g_hash_table_foreach_remove (block, (GHRFunc)free_block_item, NULL);
        g_hash_table_destroy (block);
}

static EnamlType
value_type (EnamlValue *value)
{
        if (value->block) {
                return ENAML_KEY_BLOCK;
        }
        if (value->list) {
                return ENAML_KEY_LIST;
        }
        if (value->string) {
                return ENAML_KEY_VALUE;
        }
        return ENAML_KEY_FLAG;
}

static EnamlValue *
enaml_list_lookup (const EnamlList *list, int index)
{
        GList *glist;

        glist = g_list_nth (*list, index);
        return glist ? (EnamlValue *)(glist->data) : NULL;
}

/* lookup a particular key (dot-notation is obeyed) */
static EnamlValue *
enaml_lookup (const EnamlBlock *block, const char *key)
{
        const char *dot;
        char *subkey;
        EnamlValue *value;

        dot = strchr (key, '.');
        if (dot == NULL) {
                return (EnamlValue *) g_hash_table_lookup ((GHashTable *)block, key);
        }

        subkey = g_strdup (key);
        subkey[dot-key] = 0;
        value = enaml_lookup (block, subkey);
        g_free (subkey);
        if (value && value->block) {
                /* recurse into the sub-block and continue */
                return enaml_lookup (value->block, dot+1);
        }
        while (value && value->list && dot) {
                /* assume the thing after the dot is a number */
                value = enaml_list_lookup (value->list, atoi (dot+1));
                dot = strchr (dot+1, '.');
        }
        return value;
}

static EnamlValue *
new_value (const char *blob, int length, int flags)
{
        EnamlValue *value;

        value = g_new0 (EnamlValue, 1);
        value->length = length;
        if (flags & ENAML_COPY) {
                value->string = g_malloc (length+1);
                memcpy (value->string, blob, length);
                value->string[length] = 0;
                flags |= ENAML_FREE;
        } else {
                value->string = (char *) blob;
        }

        /* only copy some flags */
        value->flags = (flags & ENAML_FREE);
        return value;
}

static EnamlValue *
new_value_list (EnamlList *list, int flags)
{
        EnamlValue *value;

        value = g_new0 (EnamlValue, 1);
        value->list = list;
        value->flags = (flags & ENAML_FREE);
        return value;
}

static EnamlValue *
new_value_block (EnamlBlock *block, int flags)
{
        EnamlValue *value;

        value = g_new0 (EnamlValue, 1);
        value->block = block;
        value->flags = (flags & ENAML_FREE);
        return value;
}

/* returns TRUE if any key was removed */
gboolean
enaml_drop_key (EnamlBlock *block, const char *key)
{
        EnamlValue *old_value;

        if (block == NULL) {
                return FALSE;
        }
        old_value = (EnamlValue *) g_hash_table_lookup (block, key);
        if (old_value == NULL) {
                return FALSE;
        } else {
                g_hash_table_remove (block, key);
                free_value (old_value);
                return TRUE;
        }
}

/* can't use dot-notation on these keys */
static void
enaml_put (EnamlBlock *block, const char *key, EnamlValue *value)
{
        strncpy (value->key, key, ENAML_MAX_KEY_SIZE);
        value->key[ENAML_MAX_KEY_SIZE-1] = 0;

        /* drop any pre-existing */
        enaml_drop_key (block, value->key);
        g_hash_table_insert (block, value->key, value);
}

static void
enaml_list_add (EnamlList *list, EnamlValue *value)
{
        *list = g_list_append (*list, value);
}

void
enaml_add_string (EnamlBlock *block, const char *key, const char *value, int flags)
{
        enaml_put (block, key, new_value (value, strlen (value), flags));
}

void
enaml_add_blob (EnamlBlock *block, const char *key, const char *value, int length,
                int flags)
{
        enaml_put (block, key, new_value (value, length, flags));
}

void
enaml_add_long (EnamlBlock *block, const char *key, long value)
{
        char *out;

        out = g_strdup_printf ("%ld", value);
        enaml_put (block, key, new_value (out, strlen (out), ENAML_FREE));
}

void
enaml_add_flag (EnamlBlock *block, const char *key)
{
        enaml_put (block, key, new_value (NULL, 0, 0));
}

void
enaml_add_list (EnamlBlock *block, const char *key, EnamlList *list, int flags)
{
        enaml_put (block, key, new_value_list (list, flags));
}

void
enaml_add_block (EnamlBlock *block, const char *key, EnamlBlock *new_block, int flags)
{
        enaml_put (block, key, new_value_block (new_block, flags));
}

void
enaml_list_add_string (EnamlList *list, const char *value, int flags)
{
        enaml_list_add (list, new_value (value, strlen (value), flags));
}

void
enaml_list_add_blob (EnamlList *list, const char *value, int length, int flags)
{
        enaml_list_add (list, new_value (value, length, flags));
}

void
enaml_list_add_long (EnamlList *list, long value)
{
        char *out;

        out = g_strdup_printf ("%ld", value);
        enaml_list_add (list, new_value (out, strlen (out), ENAML_FREE));
}

void
enaml_list_add_list (EnamlList *list, EnamlList *list_value, int flags)
{
        enaml_list_add (list, new_value_list (list_value, flags));
}

EnamlType
enaml_type (const EnamlBlock *block, const char *key)
{
        EnamlValue *value;

        value = enaml_lookup (block, key);
        if (value) {
                return value_type (value);
        } else {
                return ENAML_KEY_UNDEF;
        }
}

EnamlType
enaml_list_type (const EnamlList *list, int index)
{
        EnamlValue *value;

        value = enaml_list_lookup (list, index);
        if (value) {
                return value_type (value);
        } else {
                return ENAML_KEY_UNDEF;
        }
}

const char *
enaml_get_string (const EnamlBlock *block, const char *key, const char *default_value)
{
        EnamlValue *value;

        value = enaml_lookup (block, key);
        if (value && value->string) {
                return value->string;
        } else {
                return default_value;
        }
}

int
enaml_get_blob (const EnamlBlock *block, const char *key, const char **blob)
{
        EnamlValue *value;

        value = enaml_lookup (block, key);
        if (value && value->string) {
                *blob = value->string;
                return value->length;
        } else {
                return 0;
        }
}
                
long
enaml_get_long (const EnamlBlock *block, const char *key, long default_value)
{
        EnamlValue *value;

        value = enaml_lookup (block, key);
        if (value && value->string) {
                return atol (value->string);
        } else {
                return default_value;
        }
}

EnamlList *
enaml_get_list (const EnamlBlock *block, const char *key)
{
        EnamlValue *value;

        value = enaml_lookup (block, key);
        if (value && value->list) {
                return value->list;
        } else {
                return NULL;
        }
}

EnamlBlock *
enaml_get_block (const EnamlBlock *block, const char *key)
{
        EnamlValue *value;

        value = enaml_lookup (block, key);
        if (value && value->block) {
                return value->block;
        } else {
                return NULL;
        }
}

const char *
enaml_list_get_string (const EnamlList *list, int index, const char *default_value)
{
        EnamlValue *value;

        value = enaml_list_lookup (list, index);
        if (value && value->string) {
                return value->string;
        } else {
                return default_value;
        }
}

int
enaml_list_get_blob (const EnamlList *list, int index, char **blob)
{
        EnamlValue *value;

        value = enaml_list_lookup (list, index);
        if (value && value->string) {
                *blob = value->string;
                return value->length;
        } else {
                return 0;
        }
}

long
enaml_list_get_long (const EnamlList *list, int index, long default_value)
{
        EnamlValue *value;

        value = enaml_list_lookup (list, index);
        if (value && value->string) {
                return atol (value->string);
        } else {
                return default_value;
        }
}

EnamlList *
enaml_list_get_list (const EnamlList *list, int index)
{
        EnamlValue *value;

        value = enaml_list_lookup (list, index);
        if (value && value->list) {
                return value->list;
        } else {
                return NULL;
        }
}

EnamlList *
enaml_copy_list (const EnamlList *list)
{
        EnamlList *newlist, *sublist;
        GList *iter;
        EnamlValue *value;

        newlist = enaml_new_list ();
        for (iter = g_list_first (*list); iter; iter = g_list_next (iter)) {
                value = (EnamlValue *)(iter->data);
                if (value->list) {
                        sublist = enaml_copy_list (value->list);
                        enaml_list_add_list (newlist, sublist, ENAML_FREE);
                } else {
                        enaml_list_add_blob (newlist, value->string, value->length, ENAML_COPY);
                }
        }
        return newlist;
}

static void
enaml_copy_block_func (const char *key, EnamlValue *value, EnamlBlock *newblock)
{
        EnamlBlock *subblock;
        EnamlList *sublist;

        if (value->block) {
                subblock = enaml_copy_block (value->block);
                enaml_add_block (newblock, key, subblock, ENAML_FREE);
                return;
        }
        if (value->list) {
                sublist = enaml_copy_list (value->list);
                enaml_add_list (newblock, key, sublist, ENAML_FREE);
                return;
        }
        if (value->string) {
                enaml_add_blob (newblock, key, value->string, value->length, ENAML_COPY);
                return;
        }
        enaml_add_flag (newblock, key);
}

EnamlBlock *
enaml_copy_block (const EnamlBlock *block)
{
        EnamlBlock *newblock;

        if (block == NULL) {
                return NULL;
        }
        newblock = enaml_new ();
        g_hash_table_foreach ((GHashTable *)block, (GHFunc)enaml_copy_block_func, (void *)newblock);
        return newblock;
}

void
enaml_copy_key (const EnamlBlock *source, EnamlBlock *dest, const char *key)
{
        int len;
        const char *blob;
        EnamlBlock *subblock;
        EnamlList *sublist;

        switch (enaml_type (source, key)) {
        case ENAML_KEY_UNDEF:
                /* ignore */
                break;
        case ENAML_KEY_FLAG:
                enaml_add_flag (dest, key);
                break;
        case ENAML_KEY_VALUE:
                len = enaml_get_blob (source, key, &blob);
                enaml_add_blob (dest, key, blob, len, ENAML_COPY);
                break;
        case ENAML_KEY_LIST:
                sublist = enaml_get_list (source, key);
                enaml_add_list (dest, key, enaml_copy_list (sublist), ENAML_FREE);
                break;
        case ENAML_KEY_BLOCK:
                subblock = enaml_get_block (source, key);
                enaml_add_block (dest, key, enaml_copy_block (subblock), ENAML_FREE);
                break;
        }
}

int
enaml_list_length (const EnamlList *list)
{
        return list ? g_list_length (*list) : 0;
}

int
enaml_block_size (const EnamlBlock *block)
{
        if (block == NULL) {
                return 0;
        } else {
                return g_hash_table_size ((GHashTable *)block);
        }
}

static void
enaml_iter_func (char *key, EnamlValue *value, EnamlIter *iter)
{
        iter->list = g_list_append (iter->list, key);
}

const char *
enaml_iter_first (const EnamlBlock *block, EnamlIter *iter)
{
        iter->list = NULL;

        g_hash_table_foreach ((GHashTable *)block, (GHFunc)enaml_iter_func, (void *)iter);
        iter->item = g_list_first (iter->list);
        if (iter->item) {
                return (const char *)(iter->item->data);
        }
        return NULL;
}

const char *
enaml_iter_next (EnamlIter *iter)
{
        iter->item = g_list_next (iter->item);
        if (iter->item) {
                return (const char *)(iter->item->data);
        }
        enaml_iter_done (iter);
        return NULL;
}

void
enaml_iter_done (EnamlIter *iter)
{
        iter->item = NULL;
        if (iter->list) {
                g_list_free (iter->list);
        }
        iter->list = NULL;
}
