/*
 * ENAML's Not A Markup Language
 *
 * Copyright (C) 2000, 2001  Robey Pointer <robey@lag.net>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef _ENAML_H_
#define _ENAML_H_

#include <glib.h>

typedef GList *		EnamlList;	/* GList<EnamlValue *> */
typedef GHashTable	EnamlBlock;	/* GHashTable<char *, EnamlValue *> */

#define ENAML_MAX_KEY_SIZE	32

/* string or list */
typedef struct {
        char key[ENAML_MAX_KEY_SIZE];
        int quotes;		/* # of bytes in the val that need to be quoted */
        int flags;
        char *string;
        int length;
        EnamlList *list;
        EnamlBlock *block;
} EnamlValue;

#define ENAML_FREE	(1<<1)	/* memory should be freed by enaml */
#define ENAML_DIGITS	(1<<2)	/* string is all digits */
#define ENAML_COPY	(1<<3)	/* copy value before saving it (only strings) */

typedef enum {
        ENAML_KEY_UNDEF = 0,		/* key or index not found */
        ENAML_KEY_FLAG,
        ENAML_KEY_VALUE,
        ENAML_KEY_LIST,
        ENAML_KEY_BLOCK
} EnamlType;



EnamlBlock *enaml_new (void);
void enaml_free (EnamlBlock *block);
EnamlList *enaml_new_list (void);
void enaml_free_list (EnamlList *list);

/* remove a key from a block (freeing the value if necessary) -- returns TRUE if the key existed */
gboolean enaml_drop_key (EnamlBlock *block, const char *key);

/* functions to add a (key, value) pair to a block.
 * the key cannot use dot-notation (ie "name" not "part.name").
 * if the ENAML_FREE flag is set, you are turning the string/blob/list/block over to enaml,
 *     and enaml is responsible for freeing it later (as part of the whole block).  otherwise,
 *     you are promising that the string/blob/list/block will be around at least as long as
 *     this block will, so you remain responsible for freeing it later.
 * if the ENAML_COPY flag is set when adding a string or blob, enaml will make a copy of the
 *     string/blob before putting it in the block.  this implicitly sets ENAML_FREE.
 */
void enaml_add_string (EnamlBlock *block, const char *key, const char *value, int flags);
void enaml_add_blob (EnamlBlock *block, const char *key, const char *value, int length, int flags);
void enaml_add_long (EnamlBlock *block, const char *key, long value);
void enaml_add_flag (EnamlBlock *block, const char *key);
void enaml_add_list (EnamlBlock *block, const char *key, EnamlList *list, int flags);
void enaml_add_block (EnamlBlock *block, const char *key, EnamlBlock *new_block, int flags);

/* functions to add an item to a list.
 * only strings/blobs or other lists may be added to a list.
 * the ENAML_FREE and ENAML_COPY flags act just like they do for the enaml_add_* functions.
 */
void enaml_list_add_string (EnamlList *list, const char *value, int flags);
void enaml_list_add_blob (EnamlList *list, const char *value, int length, int flags);
void enaml_list_add_long (EnamlList *list, long value);
void enaml_list_add_list (EnamlList *list, EnamlList *list_value, int flags);

/* find out what type an item is, in a block or list */
EnamlType enaml_type (const EnamlBlock *block, const char *key);
EnamlType enaml_list_type (const EnamlList *list, int index);

/* functions to pull a value out of a block or list.
 * if the key could not be found, your default value is returned.
 * for the key, a dotted-form is allowed.  this lets you specify a sub-member of a
 * block or list (as deep as you like) with a single key.  for example, in:
 *     house:{ bedrooms:2 other:{ garage pets:"yes" } }
 * you can fetch the key "house.other.pets" to get "yes".  if you reference a list,
 * the last section of the key can be list indices:
 *     house:{ bedrooms:[ "master" "guest" ] }
 * the key "house.bedrooms.1" will return "guest".
 *
 * enaml_get_blob returns the size of the blob, or 0 if there's no data or no such key.
 */
const char *enaml_get_string (const EnamlBlock *block, const char *key, const char *default_value);
int enaml_get_blob (const EnamlBlock *block, const char *key, const char **blob);
long enaml_get_long (const EnamlBlock *block, const char *key, long default_value);
EnamlList *enaml_get_list (const EnamlBlock *block, const char *key);
EnamlBlock *enaml_get_block (const EnamlBlock *block, const char *key);
const char *enaml_list_get_string (const EnamlList *list, int index, const char *default_value);
int enaml_list_get_blob (const EnamlList *list, int index, char **blob);
long enaml_list_get_long (const EnamlList *list, int index, long default_value);
EnamlList *enaml_list_get_list (const EnamlList *list, int index);

/* find the number of keys in a block, or number of items in a list */
int enaml_block_size (const EnamlBlock *block);
int enaml_list_length (const EnamlList *list);

/* compose an enaml block into a string */
char *enaml_compose (const EnamlBlock *block, gboolean add_cr);
void enaml_compose_free (char *s);

/* list of keys that were in a block at the moment that enaml_iter_first() was called */
typedef struct {
        GList *list;
        GList *item;
} EnamlIter;

/* functions for iterating through the keys in a block.
 * enaml_iter_next() will return NULL when the keyspace is exhausted.  if you stop iterating
 * before you reach that point, you need to call enaml_iter_done() to free memory.
 */
const char *enaml_iter_first (const EnamlBlock *block, EnamlIter *iter);
const char *enaml_iter_next (EnamlIter *iter);
void enaml_iter_done (EnamlIter *iter);

/* for your convenience */
#define ENAML_FOREACH(_key, _block, _iter) \
	for (_key = enaml_iter_first (_block, &(_iter)); _key; _key = enaml_iter_next (&(_iter)))

/* deep-copies a block or list (identical block or list that shares no memory with the original) */
EnamlList *enaml_copy_list (const EnamlList *list);
EnamlBlock *enaml_copy_block (const EnamlBlock *block);

/* deep-copy a key from one block to another */
void enaml_copy_key (const EnamlBlock *source, EnamlBlock *dest, const char *key);

/* parse a string into a block (reverse of compose) */
EnamlBlock *enaml_parse (const char *string);

/* build a block out of component parts easily */
EnamlBlock *enaml_vbuildf (const char *tag, const char *fmt, va_list arg);
EnamlBlock *enaml_buildf (const char *tag, const char *fmt, ...);

#endif	/* _ENAML_H_ */
