/* $Id: log.c,v 1.6 2001/07/12 23:41:20 robey Exp $
 *
 * Logs the various messages which tinyproxy produces to either a log file or
 * the syslog daemon. Not much to it...
 *
 * Copyright (C) 1998  Steven Young
 * Copyright (C) 1999  Robert James Kaes (rjkaes@flarenet.com)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * log.c - For the manipulation of log files.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdarg.h>
#include <time.h>
#include <syslog.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <glib.h>

#include "log.h"

#define LENGTH 24

static FILE *logfp = NULL;


static void
say2_log_handler (const char *log_domain, GLogLevelFlags flags, const char *message, void *userdata)
{
        const char *typestr = "???";

        switch (flags & G_LOG_LEVEL_MASK) {
        case G_LOG_LEVEL_ERROR:
                typestr = "ERROR";
                break;
        case G_LOG_LEVEL_CRITICAL:
                typestr = "CRITICAL";
                break;
        case G_LOG_LEVEL_WARNING:
                typestr = "warning";
                break;
        case G_LOG_LEVEL_MESSAGE:
                typestr = "notice";
                break;
        case G_LOG_LEVEL_INFO:
                typestr = "info";
                break;
        case G_LOG_LEVEL_DEBUG:
                typestr = "debug";
                break;
        }

        log ("%s: %s", typestr, message);
}

static void
check_g_log_handler (void)
{
        static gboolean handler_active = FALSE;

        if (handler_active) {
                return;
        }
        g_log_set_handler (NULL, G_LOG_FLAG_RECURSION | G_LOG_FLAG_FATAL | G_LOG_LEVEL_ERROR |
                           G_LOG_LEVEL_CRITICAL | G_LOG_LEVEL_WARNING | G_LOG_LEVEL_MESSAGE,
                           say2_log_handler, NULL);
        handler_active = TRUE;
}

gboolean
log_open (const char *filename)
{
        if (logfp && (logfp != stdout) && (logfp != stderr)) {
                fclose (logfp);
        }
        if (strcmp(filename, "-") == 0) {
                /* stdout */
                logfp = stdout;
        } else {
                logfp = fopen (filename, "a");
        }
        check_g_log_handler ();
        return (logfp != NULL);
}

void
log_close (void)
{
        fclose (logfp);
        logfp = NULL;
}

void
log_set (FILE *fp)
{
        if (logfp && (logfp != fp) && (logfp != stdout) && (logfp != stderr)) {
                fclose (logfp);
        }
        logfp = fp;
        check_g_log_handler ();
}

/* write something to the log file */
void
log (const char *fmt, ...)
{
	va_list args;
	time_t nowtime;
	static char time_string[LENGTH];
	unsigned char *out, *p;

	g_assert (fmt);

	if (! logfp) {
		/* then just don't log anything */
		return;
	}

	va_start (args, fmt);

	nowtime = time (NULL);
	/* Format is day month year hour:minute:second (24 time) */
	strftime (time_string, LENGTH, "%d %b %Y %H:%M:%S", localtime (&nowtime));

	out = (unsigned char *) g_strdup_vprintf (fmt, args);
	/* weed out any questionable chars */
	for (p = out; *p; p++) {
		if ((*p < ' ') || (*p == 127)) {
			*p = '?';
		}
	}

	fprintf (logfp, "%s [%d]: %s", time_string, (int) getpid (), (char *)out);
	fprintf (logfp, "\n");
	fflush (logfp);

	g_free (out);
	va_end (args);
}
