/* $Id: sock.c,v 1.13 2001/05/03 05:55:40 robey Exp $
 *
 * Handles incoming request callbacks, buffers headers, and passes
 * the request on to the upstream server.
 *
 * [robey, 05apr2001] As an experiment, I ripped out all the code that had us
 * calling poll() directly.  Instead, we hook into the GIOChannel system, and
 * let glib do all the polling for us.  (We still use the fd's directly for
 * any read/write/connect/listen/accept stuff.)
 *
 * Copyright (C) 2000, 2001 Eazel, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Robey Pointer <robey@eazel.com>
 *          Mike Fleming <mfleming@eazel.com>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#include <glib.h>

#include "log.h"
#include "sock.h"
#include "dnscache.h"
#include "sock-ssl.h"

#define IS_TUNNEL_LEFT_RIGHT(left,right)  ((left)->in_buffer == (right)->out_buffer)
#define IS_TUNNEL_RIGHT_LEFT(left,right)  ((right)->in_buffer == (left)->out_buffer)

/* sockets we have open */
static Socket *socket_chain = NULL;

static gboolean socket_channel_callback (GIOChannel *channel, GIOCondition events, Socket *sock);


/* returns 1 on success */
static SockBuf *
sockbuf_new(void)
{
        SockBuf *buf = g_new0 (SockBuf, 1);

        buf->buffer = (char *) g_malloc (SOCKBUF_SIZE);
        buf->size = SOCKBUF_SIZE;
        buf->used = 0;
        return buf;
}

static void
sockbuf_free(SockBuf *buf)
{
        g_return_if_fail (buf != NULL);
        g_return_if_fail (buf->buffer != NULL);

        g_free (buf->buffer);
        buf->size = buf->used = 0;
        buf->buffer = NULL;
        g_free (buf);
}

/* passing 0 for target_size doubles the buffer size */
static void
sockbuf_step_size (SockBuf *buf, int target_size)
{
        int new_size;

        /* find the smallest n such that buf->size * 2^n > buf->used + datalen */
        for (new_size = buf->size * 2 ; new_size < target_size ; new_size *= 2) ;

        buf->buffer = g_realloc (buf->buffer, new_size);
        buf->size = new_size;
}

/* if there's a complete (linefeed-terminated) line in the sockbuf, it's pulled off and returned --
 * the returned string must be g_free()'d later.
 */
static char *
sockbuf_getline(SockBuf *buf)
{
	int i, end;
	char *line;

	for (i = 0; i < buf->used; i++) {
		if ((buf->buffer[i] == '\n') || (buf->buffer[i] == '\r')) {
			/* found a line */
			break;
		}
	}

	if (i >= buf->used) {
		/* no complete line yet */
                if (buf->used == buf->size) {
                        sockbuf_step_size (buf, 0);
                }
		return NULL;
	}

	end = i-1;
	/* push i past the linefeed */
	i++;
	if ((i < buf->used) && (buf->buffer[i-1] == '\r') && (buf->buffer[i] == '\n')) {
		i++;
	}

	line = (char *) g_malloc (end+2);
	if (!line) {
		/* out of memory? */
		return NULL;
	}
	memcpy (line, buf->buffer, end+1);
	line[end+1] = 0;
	memmove (buf->buffer, buf->buffer + i, buf->used - i);
	buf->used -= i;

	return line;
}

static int
sockbuf_get_data (SockBuf *buf, char ** pp_data, size_t *p_len)
{
	if (buf->used) {
		*p_len = buf->used;
		*pp_data = g_malloc (sizeof(char)*buf->used);
		memcpy (*pp_data,  buf->buffer, buf->used);
		buf->used = 0;
	} else {
		*p_len = 0;
		*pp_data = NULL;
	}

	return 1;
}


/* dump as much of a sockbuf as we can, into a socket */
/* returns 0, or an errno if an error occurred */
static int
sockbuf_write(SockBuf *buf, Socket *sock)
{
	int bytes;

	/* shouldn't have been polling for write if there was nothing in the buffer */
	g_assert (buf->used > 0);
	g_assert (sock != NULL);
	g_assert (buf != NULL);

	if (sock->flags & SOCKET_FLAG_USE_SSL) {
#ifdef HAVE_OPENSSL
		bytes = SSL_write (sock->ssl_handle, buf->buffer, buf->used);
#else
                bytes = 0;
#endif
	} else {
		bytes = write (sock->fd, buf->buffer, buf->used);
	}
	if (bytes < 0) {
		return errno;
	}

#ifdef DEBUG
	if (sock->output_mirror) {
		fwrite (buf->buffer, buf->used, 1, sock->output_mirror);
	}
#endif 

	sock->bytes_written += bytes;

	if (bytes != buf->used) {
		memmove (buf->buffer, buf->buffer + bytes, buf->used - bytes);
	}
	buf->used -= bytes;

	return 0;
}

/* read as much as we can into a sockbuf */
/* returns 0, or an errno if an error occurred, or -1 on EOF */
static int
sockbuf_read (SockBuf *buf, Socket *sock)
{
	int bytes;

	/* shouldn't have been polling for read if there was no buffer space */
	g_assert (buf->used < buf->size);
	g_assert (sock != NULL);
	g_assert (buf != NULL);

	if (sock->flags & SOCKET_FLAG_USE_SSL) {
#ifdef HAVE_OPENSSL
		bytes = SSL_read (sock->ssl_handle, buf->buffer + buf->used, buf->size - buf->used);
#else
                bytes = 0;
#endif
	} else {
		bytes = read (sock->fd, buf->buffer + buf->used, buf->size - buf->used);
	}
	if (bytes < 0) {
                if ((errno != EAGAIN) && (errno != EWOULDBLOCK)) {
                        return errno;
                } else {
                        return 0;
                }
	}
        if (bytes == 0) {
                return -1;
        }

#ifdef DEBUG
	if (sock->input_mirror) {
		fwrite (buf->buffer + buf->used, bytes, 1, sock->input_mirror);
	}
#endif

	sock->bytes_read += bytes;

	buf->used += bytes;
	return 0;
}

static int
sockbuf_append(SockBuf *buf, char *data, int datalen)
{
	if (buf->used + datalen > buf->size) {
                sockbuf_step_size (buf, buf->used + datalen);
	}
	memcpy (buf->buffer + buf->used, data, datalen);
	buf->used += datalen;
	return 1;
}

static int
sockbuf_append_buffer(SockBuf *buf, SockBuf *xbuf)
{
        sockbuf_append (buf, xbuf->buffer, xbuf->used);
        xbuf->used = 0;
        return 1;
}


/* returns TRUE on success */
static gboolean
resolve_name (const char *name, struct in_addr *addr)
{
	return (dnscache_lookup (addr, name) == 0) ? TRUE : FALSE;
}

gboolean
socket_dns_resolves (const char *name)
{
        struct in_addr addr;

        return resolve_name (name, &addr);
}

/* annoying callback required because glib defers the destruction of a channel */
static void
socket_channel_destroy (Socket *sock)
{
        Socket *s;

        if (! (sock->flags & SOCKET_FLAG_DESTROYED)) {
                return;
        }

        /* free the socket buffers */
	if (sock->tunnel) {
		/* If tunnelling was only in one direction, then one buffer
		 * needs to be disposed of.  Otherwise, the buffers
		 * will be disposed of by the other socket's socket_destroy
		 */

		if ( ! IS_TUNNEL_LEFT_RIGHT (sock, sock->tunnel)) {
	                sockbuf_free (sock->in_buffer);
		}
		if ( ! IS_TUNNEL_RIGHT_LEFT (sock, sock->tunnel)) {
	                sockbuf_free (sock->out_buffer);
		}
		
		/* disconnect any existing tunnel */
                s = sock->tunnel;
		s->tunnel = NULL;
		sock->tunnel = NULL;

		/* signal close for other socket */
                /* FIXME: ... */
                /* shutdown bug? */
                socket_close (s);
	} else {
                sockbuf_free (sock->in_buffer);
                sockbuf_free (sock->out_buffer);
        }

        if (socket_chain == sock) {
                socket_chain = sock->next;
        } else {
                for (s = socket_chain; s->next; s = s->next) {
                        if (s->next == sock) {
                                s->next = sock->next;
                                break;
                        }
                }
        }
	g_free (sock);
}
        
/* make sure this socket is in the glib poll queue, and possibly change its event flags */
static void
socket_watch (Socket *sock, GIOCondition events)
{
        if (sock->channel != NULL) {
                g_source_remove (sock->source_id);
        } else {
                sock->channel = g_io_channel_unix_new (sock->fd);
        }
        if (events != 0) {
                sock->source_id = g_io_add_watch_full (sock->channel, 0, events | G_IO_HUP,
                                                       (GIOFunc)socket_channel_callback, sock,
                                                       (GDestroyNotify)socket_channel_destroy);
        } else {
                sock->source_id = 0;
                if (sock->channel != NULL) {
                        g_io_channel_unref (sock->channel);
                        sock->channel = NULL;
		} else {
                        /* channel was already destroyed a while ago -- force cleanup */
                        socket_channel_destroy (sock);
                }
        }
}

/* figure out what events need to be trapped, and possibly change the socket watch */
static void
socket_reevaluate_events (Socket *sock)
{
        GIOCondition events = 0;

        if (sock->flags & SOCKET_FLAG_DESTROYED) {
                /* don't bother. */
                return;
        }

        /* don't do any events on a frozen socket */
        if (sock->frozen_count == 0) {
                if ((sock->flags & SOCKET_FLAG_LISTEN) ||
                    (sock->in_buffer->used < sock->in_buffer->size)) {
                        events |= G_IO_IN;
                }
                if ((sock->flags & SOCKET_FLAG_CONNECTING) ||
                    ((sock->out_buffer->used > 0) && !(sock->flags & SOCKET_FLAG_CORK))) {
                        events |= G_IO_OUT;
                }
        }
        if (events != sock->events) {
                socket_watch (sock, events);
                sock->events = events;
        }
}

void
socket_destroy (Socket *sock)
{
	g_assert (IS_SOCKET (sock));

	sock->magic = 0;

#ifdef HAVE_OPENSSL
        if (sock->ssl_cert_buffer != NULL) {
                g_free (sock->ssl_cert_buffer);
                sock->ssl_cert_buffer = NULL;
                sock->ssl_cert_length = 0;
        }
        if (sock->ssl_key_buffer != NULL) {
                g_free (sock->ssl_key_buffer);
                sock->ssl_key_buffer = NULL;
                sock->ssl_key_length = 0;
        }

        if (sock->ssl_handle != NULL) {
                socket_ssl_free (sock->ssl_handle);
                sock->ssl_handle = NULL;
        }
#endif

	if (sock->fd >= 0) {
		close (sock->fd);
		sock->fd = -1;
	}


#ifdef DEBUG
	if (sock->input_mirror) {
		fclose (sock->input_mirror);
	}
	if (sock->output_mirror) {
		fclose (sock->output_mirror);
	}
#endif

        sock->flags |= SOCKET_FLAG_DESTROYED;
        /* turn off any glib polling */
        socket_watch (sock, 0);
}

/* create a new socket.  if name is non-NULL, bind to that address */
Socket *
socket_new (const char *name, int port)
{
	Socket *sock;
	int flags, i;
	struct in_addr addr;
	struct sockaddr_in sockaddr;

        sock = g_new0 (Socket, 1);
	sock->magic = SOCKET_MAGIC;
        sock->next = socket_chain;
        socket_chain = sock;

        sock->in_buffer = sockbuf_new ();
        sock->out_buffer = sockbuf_new ();
        if (!sock->in_buffer || !sock->out_buffer) {
		socket_destroy (sock);
		return NULL;
	}

	sock->fd = socket (AF_INET, SOCK_STREAM, 0);
	if (sock->fd < 0) {
		socket_destroy (sock);
		return NULL;
	}

        /* reuse address */
        i = 1;
        if (setsockopt (sock->fd, SOL_SOCKET, SO_REUSEADDR, &i, sizeof (i)) != 0) {
                socket_destroy (sock);
                return NULL;
        }
                         
	/* set it nonblocking */
	flags = fcntl (sock->fd, F_GETFL, 0);
	fcntl (sock->fd, F_SETFL, flags | O_NONBLOCK);

	/* bind, if an address was given */
	memset (&sockaddr, 0, sizeof(struct sockaddr_in));
	memset (&addr, 0, sizeof(struct in_addr));
	sockaddr.sin_family = AF_INET;
	if (name) {
		resolve_name (name, &addr);
		memcpy (&sockaddr.sin_addr.s_addr, &addr, sizeof(struct in_addr));
	}
	sockaddr.sin_port = htons (port);

	if (bind (sock->fd, (struct sockaddr *)&sockaddr, sizeof(struct sockaddr_in)) != 0) {
		/* failed to bind. :( */
		socket_destroy (sock);
		return NULL;
	}

	return sock;
}

#ifdef DEBUG
void
socket_set_mirrors (Socket *sock, FILE *output_mirror, FILE *input_mirror)
{
	if (sock->output_mirror) {
		fclose (sock->output_mirror);
	}

	sock->output_mirror = output_mirror;

	if (sock->input_mirror) {
		fclose (sock->input_mirror);
	}

	sock->input_mirror = input_mirror;
}
#endif /* DEBUG */

void
socket_close (Socket *sock)
{
        g_assert (sock != NULL);

        if (sock->flags & SOCKET_FLAG_CLOSING) {
                return;
        }

        sock->flags |= SOCKET_FLAG_CLOSING;
        sock->time = time (NULL);

	if (sock->out_buffer->used) {
		/* still waiting for socket to write -- wait before closing */
		return;
	}

	if (sock->eoffn) {
		(*sock->eoffn) (sock);
	}
	socket_destroy (sock);
}

/* return the next line from a socket (if there is one).
 * you must g_free() it when done.
 */
char *
socket_getline (Socket *sock)
{
        char *out;

	out = sockbuf_getline (sock->in_buffer);
        socket_reevaluate_events (sock);
        return out;
}

/* write data to a socket (might be buffered for later if the socket is busy) */
int
socket_write (Socket *sock, char *data, int datalen)
{
	int bytes;
        int ret;

	g_return_val_if_fail(sock != NULL, 0);
	g_return_val_if_fail(!(sock->flags & (SOCKET_FLAG_LISTEN | SOCKET_FLAG_CONNECTING |
					      SOCKET_FLAG_CLOSING)), 0);

        if ((sock->flags & SOCKET_FLAG_CORK) || sock->out_buffer->used > 0) {
                /* either we're in cork mode or already have output queued to this socket
                 *  -- just append
                 */
		return sockbuf_append (sock->out_buffer, data, datalen);
	}

	/* not buffering, so try a direct write first */
	if (sock->flags & SOCKET_FLAG_USE_SSL) {
#ifdef HAVE_OPENSSL
		bytes = SSL_write (sock->ssl_handle, data, datalen);
#else
                bytes = 0;
#endif
	} else {
		bytes = write (sock->fd, data, datalen);
	}

	if (bytes < 0) {
		return 0;	/* ouch */
	}

#ifdef DEBUG
	if (sock->output_mirror && bytes > 0) {
		fwrite (data, bytes, 1, sock->output_mirror); 
	}
#endif
	sock->bytes_written += bytes;

	if (bytes == datalen) {
		return 1;
	}

	/* wrote less than we had -- buffer the rest */
	ret = sockbuf_append (sock->out_buffer, data + bytes, datalen - bytes);
        socket_reevaluate_events (sock);
        return ret;
}

int
socket_read (Socket *sock, char ** pp_data, size_t *p_len)
{
        int ret;

	g_return_val_if_fail (NULL != sock, 0);
	g_return_val_if_fail (IS_SOCKET(sock), 0);
	g_return_val_if_fail (NULL != pp_data, 0);
	g_return_val_if_fail (NULL != p_len, 0);

	ret = sockbuf_get_data (sock->in_buffer, pp_data, p_len);
        socket_reevaluate_events (sock);
        return ret;
}


static void
socket_finish_connect (Socket *sock)
{
        int err;
        int len;
	int success;

        sock->flags &= ~SOCKET_FLAG_CONNECTING;

	len = sizeof(int);
        if ((getsockopt (sock->fd, SOL_SOCKET, SO_ERROR, &err, &len) != 0) ||
            (err != 0)) {
		success = 0;
		goto out;
        }
	if (sock->flags & SOCKET_FLAG_USE_SSL) {
#ifdef HAVE_OPENSSL
		int flags;

		/* FIXME: ssl negotiation has to be done over a blocking
		 * socket.  that's kinda lame.
		 */
		flags = fcntl(sock->fd, F_GETFL, 0);
		fcntl(sock->fd, F_SETFL, flags & ~O_NONBLOCK);

		/* negotiate ssl over this connection */
                err = socket_ssl_begin_ssl (sock, sock->fd, SOCKET_SSL_CLIENT);
		if (err != 0) {
			success = 0;
			goto out;
		}
#else
                success = 0;
                goto out;
#endif
	}
	success = 1;

out:
        if (success) {
                socket_reevaluate_events (sock);
        } else {
                socket_watch (sock, 0);
        }
	if (sock->connectfn) {
		(*sock->connectfn) (sock, success);
		sock->connectfn = NULL;
	}
	return;
}

/* connect to a remote host & port.
 * returns 0 if the DNS failed, 1 if the connection is being attempted
 */
int
socket_connect (Socket *sock, const char *name, int port, SocketConnectFn connectfn)
{
        struct sockaddr_in sockaddr;
        struct in_addr addr;
        int err;

	g_return_val_if_fail (sock != NULL, 0);

	memset (&sockaddr, 0, sizeof(struct sockaddr_in));
	memset (&addr, 0, sizeof(struct in_addr));
	sockaddr.sin_family = AF_INET;
        if (! resolve_name (name, &addr)) {
                return 0;
        }
        memcpy (&sockaddr.sin_addr.s_addr, &addr, sizeof(struct in_addr));
        sockaddr.sin_port = htons (port);

        sock->connectfn = connectfn;

        err = connect (sock->fd, (struct sockaddr *)&sockaddr, sizeof(struct sockaddr_in));
        if (err == 0) {
                /* instant connect! */
		socket_finish_connect (sock);
                return 1;
        }
        if (errno == EINPROGRESS) {
                sock->flags |= SOCKET_FLAG_CONNECTING;
                socket_reevaluate_events (sock);
                return 1;
        }
        /* instant failure */
        socket_finish_connect (sock);
        return 1;
}

int
socket_connect_ssl (Socket *sock, const char *name, int port, SocketConnectFn connectfn)
{
	g_return_val_if_fail (sock != NULL, 0);

	sock->flags |= SOCKET_FLAG_USE_SSL;
	return socket_connect (sock, name, port, connectfn);
}

static void
socket_accept (Socket *sock)
{
        int newfd;
        int addrlen;
        struct sockaddr_in sockaddr;
        Socket *newsock;
	const char *host_ip;
	int port;

        addrlen = sizeof(struct sockaddr_in);
        newfd = accept (sock->fd, (struct sockaddr *)&sockaddr, &addrlen);
        if (newfd < 0) {
                return;
        }

	newsock = g_new0 (Socket, 1);
	newsock->magic = SOCKET_MAGIC;

        newsock->fd = newfd;
        newsock->next = socket_chain;
        socket_chain = newsock;
        newsock->in_buffer = sockbuf_new ();
        newsock->out_buffer = sockbuf_new ();
        if (!newsock->in_buffer || !newsock->out_buffer) {
		socket_destroy (newsock);
		return;
	}

#ifdef HAVE_OPENSSL
        /* vitally important that we copy any certs/pkeys over to the new socket */
        if (sock->ssl_cert_length > 0) {
                newsock->ssl_cert_length = sock->ssl_cert_length;
		newsock->ssl_cert_buffer = g_malloc (sock->ssl_cert_length);
                memcpy (newsock->ssl_cert_buffer, sock->ssl_cert_buffer, sock->ssl_cert_length);
        }
        if (sock->ssl_key_length > 0) {
                newsock->ssl_key_length = sock->ssl_key_length;
                newsock->ssl_key_buffer = g_malloc (sock->ssl_key_length);
                memcpy (newsock->ssl_key_buffer, sock->ssl_key_buffer, sock->ssl_key_length);
        }
#endif

        socket_reevaluate_events (newsock);

        if (sock->acceptfn) {
                dnscache_lookup_reverse (&sockaddr.sin_addr, &host_ip);
		port = ntohs (sockaddr.sin_port);
                (*sock->acceptfn) (newsock, sock->data, host_ip, port);
        }
}

static void
socket_eof (Socket *sock)
{
        if (sock->flags & SOCKET_FLAG_CLOSING) {
                return;
        }
        sock->flags |= SOCKET_FLAG_CLOSING;

        if (sock->eoffn) {
                (*sock->eoffn) (sock);
        }
        socket_destroy (sock);
}

int
socket_listen (Socket *sock, SocketAcceptFn acceptfn)
{
        if (listen (sock->fd, 64) != 0) {
                return 0;
        }

        sock->flags |= SOCKET_FLAG_LISTEN;
        sock->acceptfn = acceptfn;
        socket_reevaluate_events (sock);

        return 1;
}

static gboolean
socket_channel_callback (GIOChannel *channel, GIOCondition events, Socket *sock)
{
        int err;

        g_assert (!(sock->flags & SOCKET_FLAG_DESTROYED));

        if (events & G_IO_OUT) {
                if (sock->flags & SOCKET_FLAG_CONNECTING) {
                        socket_finish_connect (sock);
                        goto out;
                } else {
                        sockbuf_write (sock->out_buffer, sock);
                        if (!(sock->out_buffer->used) && (sock->flags & SOCKET_FLAG_CLOSING)) {
                                /* can finally close this socket */
                                sock->flags &= ~SOCKET_FLAG_CLOSING;
                                socket_close (sock);
                                goto out;
                        }
                        if (! sock->out_buffer->used) {
                                /* stop polling for write */
                                socket_reevaluate_events (sock);
                        }
                }
        }

        if (events & G_IO_IN) {
                if (sock->flags & SOCKET_FLAG_LISTEN) {
                        socket_accept (sock);
                        goto out;
                } else if (sock->flags & SOCKET_FLAG_CONNECTING) {
                        socket_finish_connect (sock);
                        goto out;
                } else {
                        err = sockbuf_read (sock->in_buffer, sock);
                        if (err != 0) {
                                socket_eof (sock);
                                goto out;
                        } else if (sock->readfn) {
                                (*sock->readfn) (sock);
                                if (sock->in_buffer->used == sock->in_buffer->size) {
                                        /* stop polling for read */
                                        socket_reevaluate_events (sock);
                                }
                                goto out;
                        }
                }
        }

        if ((events & G_IO_HUP) || (events & G_IO_ERR)) {
                if (sock->flags & SOCKET_FLAG_CONNECTING) {
                        socket_finish_connect (sock);
                        goto out;
                } else {
                        socket_eof (sock);
                        goto out;
                }
        }

out:
        /* always return TRUE -- if we need to remove the source, that'll be done in the callbacks */
        return TRUE;
}

/**
 * socket_freeze
 * 
 * Temporarally prevent socket events on this particular from being processed,
 * and thus generating callbacks.  Useful for modal event loops.
 */

void
socket_freeze (Socket *sock)
{
        sock->frozen_count++;
        socket_reevaluate_events (sock);
}

/**
 * socket_thaw
 * 
 * undo a socket_freeze
 */

void
socket_thaw (Socket *sock)
{
        sock->frozen_count--;
        g_assert ( sock->frozen_count >= 0 );
        socket_reevaluate_events (sock);
}

/**
 * socket_tunnel
 * 
 * Connect two sockets directly together, directing the output of one into another
 * 
 * socket_tunnel can set up tunnelling in one direction or both directions.  Once
 * tunnelling has been set up one direction, socket_tunnel can be called again
 * to establish it in both directions but it can never be disabled in a direction
 * once it has been established.
 * 
 * Once tunnelling has been established between two sockets, neither of 
 * the sockets may be tunneled to any other socket.
 */

void
socket_tunnel (Socket *left, Socket *right, SocketTunnelMode mode)
{
        g_return_if_fail (left != NULL);
        g_return_if_fail (right != NULL);
	g_return_if_fail (left != right);
        /* Can only tunnel to a socket you've already tunneled to */
        g_return_if_fail (left->tunnel == NULL || left->tunnel == right);
        g_return_if_fail (right->tunnel == NULL || right->tunnel == left);
	/* Can only upgrade modes--can't downgrade */
	g_return_if_fail ( !(IS_TUNNEL_LEFT_RIGHT(left,right) && !(mode & SOCKET_TUNNEL_LEFT_TO_RIGHT) ) );
	g_return_if_fail ( !(IS_TUNNEL_RIGHT_LEFT(left,right) && !(mode & SOCKET_TUNNEL_RIGHT_TO_LEFT) ) );

        left->tunnel = right;
        right->tunnel = left;

	/*
	 * Tunnelling:
	 * 1. Anything existing in the tunneler's output buffer will be appended
	 *    to the tunnellee's input buffer
	 * 2. The tunneler's output buffer is disposed of, and connected directly
	 *    to the tunnellee's input buffer
	 * 3. See socket_destroy for special-case socket closing code
	 */

	if ((mode & SOCKET_TUNNEL_LEFT_TO_RIGHT) && !IS_TUNNEL_LEFT_RIGHT (left,right) ) {
	        sockbuf_append_buffer (right->out_buffer, left->in_buffer);
	        sockbuf_free (left->in_buffer);
	        left->in_buffer = right->out_buffer;
	}

	if ((mode & SOCKET_TUNNEL_RIGHT_TO_LEFT) && !IS_TUNNEL_RIGHT_LEFT (left,right) ) {
	        sockbuf_append_buffer (left->out_buffer, right->in_buffer);
	        sockbuf_free (right->in_buffer);
	        right->in_buffer = left->out_buffer;
	}
}

/* Begins SSL on an established Socket */
/* This call will block */
int
socket_begin_ssl (Socket *sock, int which)
{
#ifdef HAVE_OPENSSL
	int fcntl_flags;
	int err;
	
	sock->flags |= SOCKET_FLAG_USE_SSL;

	/* FIXME: ssl negotiation has to be done over a blocking
	 * socket.  that's kinda lame.
	 */
	fcntl_flags = fcntl (sock->fd, F_GETFL, 0);
	fcntl (sock->fd, F_SETFL, fcntl_flags & ~O_NONBLOCK);

        err = socket_ssl_begin_ssl (sock, sock->fd, which);
        if (err != 0) {
                sock->flags &= ~SOCKET_FLAG_USE_SSL;
        }

	return (err == 0);
#else
        /* always fail with no SSL */
        return 0;
#endif  /* HAVE_OPENSSL */
}

gboolean
socket_is_secure (Socket *sock)
{
#ifdef HAVE_OPENSSL
        return (sock->flags & SOCKET_FLAG_USE_SSL);
#else
        return FALSE;
#endif	/* HAVE_OPENSSL */
}

void
socket_set_read_fn (Socket *sock, SocketReadFn readfn)
{
        g_return_if_fail (sock != NULL);

	sock->readfn = readfn;
}

void
socket_set_eof_fn (Socket *sock, SocketEofFn eoffn)
{
        g_return_if_fail (sock != NULL);

	sock->eoffn = eoffn;
}

void
socket_set_data (Socket *sock, void *data)
{
        g_return_if_fail (sock != NULL);

	sock->data = data;
}

void *
socket_get_data (Socket *sock)
{
        g_return_val_if_fail (sock != NULL, NULL);

	return sock->data;
}

void
socket_cork (Socket *sock)
{
        sock->flags |= SOCKET_FLAG_CORK;
        socket_reevaluate_events (sock);
}

void
socket_uncork (Socket *sock)
{
        sock->flags &= ~SOCKET_FLAG_CORK;
        socket_reevaluate_events (sock);
}

void
socket_get_byte_counts(Socket *sock, guint32 *read, guint32 *written)
{
	g_assert (IS_SOCKET (sock));

	if (read) {
		*read = sock->bytes_read;
	}
	if (written) {
		*written = sock->bytes_written;
	}
}

int
socket_printf (Socket *sock, char *format, ...)
{
        va_list args;
        char *out;
        int ret;

        g_assert (IS_SOCKET (sock));
        g_return_val_if_fail (format != NULL, 0);

        va_start (args, format);
        out = g_strdup_vprintf (format, args);
        ret = socket_write (sock, out, strlen (out));

        g_free (out);
        va_end (args);
        return ret;
}

void
socket_set_cert (Socket *sock, const char *buffer, int length)
{
#ifdef HAVE_OPENSSL
        if (sock->ssl_cert_buffer != NULL) {
                g_free (sock->ssl_cert_buffer);
                sock->ssl_cert_buffer = NULL;
                sock->ssl_cert_length = 0;
        }
        if (buffer != NULL) {
                sock->ssl_cert_buffer = g_malloc (length);
                memcpy (sock->ssl_cert_buffer, buffer, length);
                sock->ssl_cert_length = length;
        }
#endif
}

void
socket_get_cert (Socket *sock, const char **buffer, int *length)
{
#ifdef HAVE_OPENSSL
        if (sock->ssl_cert_buffer != NULL) {
                *buffer = sock->ssl_cert_buffer;
                *length = sock->ssl_cert_length;
        } else {
#endif
                *buffer = NULL;
                *length = 0;
#ifdef HAVE_OPENSSL
        }
#endif
}

void
socket_set_private_key (Socket *sock, const char *buffer, int length)
{
#ifdef HAVE_OPENSSL
        if (sock->ssl_key_buffer != NULL) {
                g_free (sock->ssl_key_buffer);
                sock->ssl_key_buffer = NULL;
                sock->ssl_key_length = 0;
        }
        if (buffer != NULL) {
                sock->ssl_key_buffer = g_malloc (length);
                memcpy (sock->ssl_key_buffer, buffer, length);
                sock->ssl_key_length = length;
        }
#endif
}

/* returns a GList<char *> of lines of socket debug info */
GList *
socket_get_debug (void)
{
	Socket *sock;
	char *state, statebuf[10];
	char calls[20], *callptr;
        GList *outlist = NULL;
        char *line;
        time_t now;
        gboolean any_destroyed = FALSE;
        gboolean any_ssl = FALSE;

        now = time (NULL);
	for (sock = socket_chain; sock; sock = sock->next) {
		if (sock->flags & SOCKET_FLAG_LISTEN) {
			state = "lstn";
		} else if (sock->flags & SOCKET_FLAG_CONNECTING) {
			state = "conn";
                } else if (sock->tunnel) {
                        state = statebuf;
                        sprintf (statebuf, "->%2d", sock->tunnel->fd);
		} else {
			state = " -- ";
		}

		callptr = calls;
		*callptr++ = (sock->acceptfn ? 'A' : ' ');
		*callptr++ = (sock->eoffn ? 'E' : ' ');
		*callptr++ = (sock->readfn ? 'R': ' ');
		*callptr++ = (sock->connectfn ? 'C' : ' ');

		if (sock->flags & SOCKET_FLAG_USE_SSL) {
                        any_ssl = TRUE;
                }
                if (sock->flags & SOCKET_FLAG_DESTROYED) {
                        any_destroyed = TRUE;
                }

		*callptr = 0;

		line = g_strdup_printf ("%c%3d%c%s (r:%6d/w:%6d) g:%d%c%c %08X %s",
                                        sock->flags & SOCKET_FLAG_DESTROYED ? '*' : ' ',
                                        sock->fd,
                                        sock->flags & SOCKET_FLAG_USE_SSL ? '+' : ' ',
                                        state, sock->bytes_read, sock->bytes_written,
                                        sock->source_id,
                                        sock->events & G_IO_IN ? '<' : ' ',
                                        sock->events & G_IO_OUT ? '>' : ' ',
                                        (int)sock->data, calls);
                outlist = g_list_prepend (outlist, line);
		if (sock->in_buffer->used || sock->out_buffer->used) {
			line = g_strdup_printf ("        (inbuf:%5d) (outbuf:%5d)", sock->in_buffer->used,
                                                sock->out_buffer->used);
                        outlist = g_list_prepend (outlist, line);
		}
                if (sock->flags & SOCKET_FLAG_CLOSING) {
                        line = g_strdup_printf ("        closing (waited %d sec so far)",
                                                (int)((long)now - (long)sock->time));
                        outlist = g_list_prepend (outlist, line);
                }
                if (sock->flags & SOCKET_FLAG_CORK) {
                        outlist = g_list_prepend (outlist, g_strdup ("        corked"));
                }
                if (sock->frozen_count != 0) {
                        outlist = g_list_prepend (outlist, g_strdup_printf ("        frozen: %d", sock->frozen_count));
                }
#ifdef HAVE_OPENSSL
                if (sock->ssl_cert_length && sock->ssl_cert_buffer) {
                        outlist = g_list_prepend (outlist, g_strdup_printf ("        attached cert (%d bytes)",
                                                                            sock->ssl_cert_length));
                }
                if (sock->ssl_key_length && sock->ssl_key_buffer) {
                        outlist = g_list_prepend (outlist, g_strdup_printf ("        attached private key (%d bytes)",
                                                                            sock->ssl_key_length));
                }
#endif
	}
        if (any_destroyed) {
                outlist = g_list_prepend (outlist, g_strdup("* = destructed socket"));
        }
        if (any_ssl) {
                outlist = g_list_prepend (outlist, g_strdup("+ = secure connection"));
        }

        outlist = g_list_reverse (outlist);
        return outlist;
}

void
socket_free_debug (GList *list)
{
        GList *iter;

        for (iter = g_list_first (list); iter != NULL; iter = g_list_next (iter)) {
                g_free ((char *)iter->data);
        }
        g_list_free (list);
}

void
socket_log_debug (void)
{
        GList *debug_log;
        GList *iter;

	log ("::: socket debug dump :::");

        debug_log = socket_get_debug ();
        for (iter = g_list_first (debug_log); iter != NULL; iter = g_list_next (iter)) {
                log ("%s", (char *)iter->data);
        }
        socket_free_debug (debug_log);

	log ("::: end of socket debug dump :::");
}
