/* $Id: sql.c,v 1.10 2001/05/28 07:48:33 robey Exp $
 *
 * Abstractize SQL queries.
 *
 * Copyright (C) 2000  Robey Pointer <robey@lag.net>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <glib.h>
#include "log.h"
#include "sql.h"
#ifdef HAVE_MYSQL
#include <mysql/mysqld_error.h>
#include <mysql/errmsg.h>
#endif	/* HAVE_MYSQL */


const char *
sql_error_string (SQLError err)
{
        switch (err) {
        case SQL_OK:
                return "success";
        case SQL_SLEEPING:
                return "server is down; sleeping";
        case SQL_EXISTS:
                return "entry already exists";
        case SQL_INTERNAL:
                return "internal libsay2 error";
        case SQL_OTHER:
                return "unknown other SQL error";
        }
        return "???";
}

static void
sql_free_list_foreach (char **row)
{
        int i;

        for (i = 0; row[i] != NULL; i++) {
                g_free (row[i]);
        }
        g_free (row);
}

void
sql_free_list (GList *list)
{
        g_list_foreach (list, (GFunc) sql_free_list_foreach, NULL);
        g_list_free (list);
}

/* quote a string for sql.
 * sadly, mysql is abyssmal at handling odd chars, so we just quote everything.
 */
char *
sql_quote_buffer (const char *in, int len)
{
        static const char *HEX = "0123456789ABCDEF";
        char *out;
        int max, i, j;

        if (in == NULL) {
                out = g_strdup ("NULL");
                return out;
        }

        max = len;
        for (j = 0; j < len; j++) {
                if ((in[j] == '\'') || (in[j] == '%') || (in[j] < ' ') || (in[j] >= 127)) {
                        max += 2;
                }
        }
        out = g_malloc (max+3);
	i = 0;
        out[i++] = '\'';
        for (j = 0; j < len; j++) {
                if ((in[j] == '\'') || (in[j] == '%') || (in[j] < ' ') || (in[j] >= 127)) {
                        out[i++] = '%';
                        out[i++] = HEX[in[j]/16];
                        out[i++] = HEX[in[j]%16];
                } else {
                        out[i++] = in[j];
                }
        }
        out[i++] = '\'';
        out[i] = 0;
        return out;
}

char *
sql_quote_string (const char *in)
{
        return sql_quote_buffer (in, (in == NULL ? 0 : strlen (in)));
}

char *
sql_unquote_buffer (const char *in, int *len)
{
        unsigned char *out;
        unsigned const char *p;
        int i;

        if (in == NULL) {
                return NULL;
        }

        out = g_malloc (strlen (in) + 1);
        for (p = (unsigned char *)in, i = 0; *p; p++) {
                if ((*p == '%') && (*(p+1) != 0) && (*(p+2) != 0)) {
                        p++;
                        out[i] = (unsigned char) (*p >= 'A' ? *p-'A'+10 : *p-'0') * 16;
                        p++;
                        out[i] += (unsigned char) (*p >= 'A' ? *p-'A'+10 : *p-'0');
                        i++;
                } else {
                        out[i++] = *p;
                }
        }
        out[i] = 0;
        *len = i;

        return (char *)out;
}

char *
sql_unquote_string (const char *in)
{
        int len;

        return sql_unquote_buffer (in, &len);
}

/* for now, only mysql is supported */
#ifdef HAVE_MYSQL

static gboolean
retry_connection (SQLState *state)
{
        state->last_tried = time (NULL);
        state->connection = mysql_real_connect (&state->mysql, state->host, state->user,
                                                state->pass, state->db, 0, NULL, 0);
        if (state->connection != NULL) {
                log ("SQL: reconnect successful");
                return TRUE;
        } else {
                return FALSE;
        }
}

SQLState *
sql_init (const char *host, const char *user, const char *pass, const char *db)
{
        SQLState *state;

        state = g_new0 (SQLState, 1);
        state->host = g_strdup (host);
        state->user = g_strdup (user);
        state->pass = g_strdup (pass);
        state->db = g_strdup (db);
        state->last_tried = time (NULL);
        mysql_init (&state->mysql);
        state->connection = mysql_real_connect (&state->mysql, host, user, pass, db,
                                                0, NULL, 0);
        if (state->connection == NULL) {
                /* allow them to try again later */
                log ("SQL: initial connect to %s failed :(", host);
                return state;
        }
        log ("SQL: established connection to %s", host);
        return state;
}

/* convert mysql error code into easier one */
static SQLError
sql_error (int err)
{
        switch (err) {
        case 0:
                return SQL_OK;
        case ER_DUP_ENTRY:
                return SQL_EXISTS;
        default:
                return SQL_OTHER;
        }
}

static SQLError
sql_do_query (SQLState *state, const char *query)
{
        int res, err;
        const char *errstr;

        if (state->connection == NULL) {
                if (time (NULL) - state->last_tried < WAIT_BETWEEN_TRIES) {
                        /* don't even retry yet */
                        return SQL_SLEEPING;
                } else {
                        log ("SQL: trying server %s again", state->host);
                        if (! retry_connection (state)) {
                                return SQL_SLEEPING;
                        }
                }
        }
        res = mysql_query (state->connection, query);
        if (res != 0) {
                err = mysql_errno (state->connection);
                errstr = mysql_error (state->connection);
                if ((err >= CR_MIN_ERROR) && (err <= CR_MAX_ERROR)) {
                        log ("SQL: lost server %s; attempting to reconnect", state->host);
                        mysql_close (state->connection);
                        if (retry_connection (state)) {
                                res = mysql_query (state->connection, query);
                                if (res != 0) {
                                        err = mysql_errno (state->connection);
                                        errstr = mysql_error (state->connection);
                                }
                        } else {
                                err = mysql_errno (&state->mysql);
                                errstr = mysql_error (&state->mysql);
                        }
                }
        }
        if (res != 0) {
                /* you had your chance to retry. grr! */
                log ("SQL: error %d: %s", err, errstr);
                return sql_error (err);
        }

        return SQL_OK;
}

SQLError
sql_query (SQLState *state, const char *query, GList **out)
{
        MYSQL_RES *result;
        SQLError err;
        int rows, fields, i, j;
        MYSQL_ROW row;
        char **row_copy;

        *out = NULL;
        err = sql_do_query (state, query);
        if (err != SQL_OK) {
                return err;
        }

        result = mysql_store_result (state->connection);
        if (result == NULL) {
                return SQL_INTERNAL;
        }

        fields = (int) mysql_num_fields (result);
        rows = (int) mysql_num_rows (result);

        for (i = 0; i < rows; i++) {
                row = mysql_fetch_row (result);
                if (row == NULL) {
                        continue;
                }
                row_copy = g_new0 (char *, fields+1);
                for (j = 0; j < fields; j++) {
                        row_copy[j] = g_strdup (row[j]);
                }
                row_copy[fields] = NULL;
                *out = g_list_prepend (*out, row_copy);
        }
        *out = g_list_reverse (*out);

        if (result != NULL) {
                mysql_free_result (result);
        }
        return SQL_OK;
}

SQLError
sql_update (SQLState *state, const char *query, int *rows)
{
        SQLError err;

        err = sql_do_query (state, query);
        if (err != SQL_OK) {
                *rows = 0;
                log ("SQL UPDATE FAILED: %s", sql_error_string (err));
                return err;
        }
        *rows = (int) mysql_affected_rows (state->connection);

        return SQL_OK;
}             

void
sql_shutdown (SQLState *state)
{
        if (state->connection != NULL) {
                mysql_close (state->connection);
        }
        g_free (state->host);
        g_free (state->user);
        g_free (state->pass);
        g_free (state->db);
}


#else	/* HAVE_MYSQL */

SQLState *
sql_init (const char *host, const char *user, const char *pass, const char *db)
{
        return NULL;
}

SQLError
sql_query (SQLState *state, const char *query, GList **out)
{
        return SQL_INTERNAL;
}

SQLError
sql_update (SQLState *state, const char *query, int *rows)
{
        return SQL_INTERNAL;
}

void
sql_shutdown (SQLState *state)
{
        /* nothing */
}

#endif	/* HAVE_MYSQL */
