/* $Id: test-enaml.c,v 1.2 2001/04/06 04:56:09 robey Exp $
 *
 * Simple tests for trying out the enaml library.
 *
 * Copyright (C) 2000  Robey Pointer <robey@lag.net>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * tests:
 *  1. simple string composition
 *  2. simple long/flag composition
 *  3. list composition
 *  4. block composition
 *  5. nested list composition
 *  6. complex block composition
 *  7. element type checking
 *  8. key iteration
 *  9. block copying
 * 10. simple parsing
 * 11. list size checking
 * 12. parser recursion
 * 13. simple fetching
 * 14. depth fetching
 * 15. block fetching
 * 16. simple buildf
 * 17. complex buildf
 *
 * TODO:
 * make a simpler parse test
 * test %!B in buildf
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <string.h>
#include "enaml.h"


#define VERIFY(_num, _msg) do { \
	char *out = enaml_compose (block, 0); \
	if (g_strcasecmp (out, expect) != 0) { \
		fprintf (stderr, "\nTest %d failed: %s\n", _num, _msg); \
		fprintf (stderr, "\t'%s'\n!=\t'%s'\n", out, expect); \
		result = 0; \
	} else { \
		fprintf (stderr, "."); \
		fflush (stderr); \
		result = 1; \
	} \
	enaml_compose_free (out); \
} while (0)

#define FAIL(_num, _msg, _reason) do { \
	fprintf (stderr, "\nTest %d failed: %s\n", _num, _msg); \
	fprintf (stderr, "\t%s\n", _reason); \
	return 0; \
} while (0)

#define PASS do { \
	fprintf (stderr, "."); \
	fflush (stderr); \
	return 1; \
} while (0)

static int
test1 (void)
{
        EnamlBlock *block;
        const char *expect = "power:\"75%25\" name:\"Squirtle\" image:%FF004A50470303 ";
        char *copy = g_strdup ("\xff\x01JPG\x03\x03:(");
        int result;

        block = enaml_new ();
        enaml_add_string (block, "name", "Squirtle", 0);
        enaml_add_string (block, "power", "75%", 0);
        copy[1] = 0;
        enaml_add_blob (block, "image", copy, 7, ENAML_COPY);
        g_free (copy);

        VERIFY (1, "simple string composition");

        enaml_free (block);
        return result;
}

static int
test2 (void)
{
        EnamlBlock *block;
        const char *expect = "weight:12984 no-turtles scale:\"-3\" ";
        int result;

        block = enaml_new ();
        enaml_add_long (block, "weight", 12984);
        enaml_add_long (block, "scale", -3);
        enaml_add_flag (block, "no-turtles");

        VERIFY (2, "simple long/flag composition");

        enaml_free (block);
        return result;
}

static int
test3 (void)
{
        EnamlBlock *block;
        EnamlList *list;
        const char *expect = "count:[ \"first\" \"second\" 3 ] ";
        int result;

        list = enaml_new_list ();
        enaml_list_add_string (list, "first", 0);
        enaml_list_add_blob (list, "secondus", 6, 0);
        enaml_list_add_long (list, 3);

        block = enaml_new ();
        enaml_add_list (block, "count", list, ENAML_FREE);

        VERIFY (3, "list composition");

        enaml_free (block);
        return result;
}

static int
test4 (void)
{
        EnamlBlock *block, *block2;
        const char *expect = "decision:{ left:\"yes\" right:\"no!!\" } ";
        int result;

        block = enaml_new ();
        block2 = enaml_new ();
        enaml_add_string (block2, "left", "yes", 0);
        enaml_add_string (block2, "right", "no!!", 0);
        enaml_add_block (block, "decision", block2, ENAML_FREE);

        VERIFY (4, "block composition");

        enaml_free (block);
        return result;
}

static int
test5 (void)
{
        EnamlBlock *block;
        EnamlList *list1, *list2;
        const char *expect = "large:[ \"mint\" [ 23 5004 ] 7 ] small:[ 23 5004 ] ";
        int result;

        list1 = enaml_new_list ();
        enaml_list_add_long (list1, 23);
        enaml_list_add_long (list1, 5004);

        list2 = enaml_new_list ();
        enaml_list_add_string (list2, "mint", 0);
        enaml_list_add_list (list2, list1, ENAML_FREE);
        enaml_list_add_long (list2, 7);

        block = enaml_new ();
        enaml_add_list (block, "small", list1, 0);
        enaml_add_list (block, "large", list2, ENAML_FREE);

	VERIFY (5, "nested list composition");

        enaml_free (block);
        return result;
}

static int
test6 (void)
{
        EnamlBlock *block, *block2, *block3, *block4;
        EnamlList *list;
        const char *expect = "begin-weird:{ other:{ empty } count:8 stats:{ weight:\"190#\" height:\"6%274%22\" } } random:[ \"clock\" \"tower\" ] ";
        int result;

        block4 = enaml_new ();
        enaml_add_string (block4, "height", "6'4\"", 0);
        enaml_add_string (block4, "weight", "190#", 0);

        block3 = enaml_new ();
        enaml_add_flag (block3, "empty");

        block2 = enaml_new ();
        enaml_add_long (block2, "count", 8);
        enaml_add_block (block2, "stats", block4, ENAML_FREE);
        enaml_add_block (block2, "other", block3, 0);

        list = enaml_new_list ();
        enaml_list_add_string (list, "clock", 0);
        enaml_list_add_string (list, "tower", 0);

        block = enaml_new ();
        enaml_add_block (block, "begin-weird", block2, ENAML_FREE);
        enaml_add_list (block, "random", list, ENAML_FREE);

        VERIFY (6, "complex block composition");

        enaml_free (block3);
        enaml_free (block);
        return result;
}

static int
test7 (void)
{
        EnamlBlock *block, *block2;
        EnamlList *list;
        const char *test = "element type checking";

        block2 = enaml_new ();
        enaml_add_flag (block2, "flag");

        list = enaml_new_list ();
        enaml_list_add_long (list, 90);

        block = enaml_new ();
        enaml_add_string (block, "string", "tigger", 0);
        enaml_add_block (block, "block", block2, ENAML_FREE);
        enaml_add_list (block, "list", list, ENAML_FREE);

        if (enaml_type (block2, "flag") != ENAML_KEY_FLAG) {
                FAIL (7, test, "flag type broken");
        }
        if (enaml_type (block, "string") != ENAML_KEY_VALUE) {
                FAIL (7, test, "string type broken");
        }
        if (enaml_type (block, "list") != ENAML_KEY_LIST) {
                FAIL (7, test, "list type broken");
        }
        if (enaml_type (block, "block") != ENAML_KEY_BLOCK) {
                FAIL (7, test, "block type broken");
        }
        if (enaml_type (block, "undef") != ENAML_KEY_UNDEF) {
                FAIL (7, test, "undef type broken");
        }

        enaml_free (block);
        PASS;
}

static int
test8 (void)
{
        EnamlBlock *block;
	EnamlIter iter;
        const char *key;
        int got_first = 0, got_second = 0, got_third = 0, unknown = 0;

        block = enaml_new ();
        enaml_add_flag (block, "first");
        enaml_add_flag (block, "second");
        enaml_add_flag (block, "third");

        ENAML_FOREACH (key, block, iter) {
                if (strcmp (key, "first") == 0) {
                        got_first = 1;
                } else if (strcmp (key, "second") == 0) {
                        got_second = 1;
                } else if (strcmp (key, "third") == 0) {
                        got_third = 1;
                } else {
                        unknown++;
                }
        }

        enaml_free (block);
        if (got_first && got_second && got_third && !unknown) {
                PASS;
        } else {
                FAIL (8, "key iteration", "missing or extra keys");
        }
}

static int
test9 (void)
{
        EnamlBlock *block, *block_orig, *block2, *block3;
        EnamlList *list;
        const char *expect = "begin:{ name:\"ack\" list:[ \"hello\" 42 ] block:{ port:6667 id:%00000017 } } warning ";
        int result;

        list = enaml_new_list ();
        enaml_list_add_string (list, "hello", 0);
        enaml_list_add_long (list, 42);

        block3 = enaml_new ();
        enaml_add_long (block3, "port", 6667);
        enaml_add_blob (block3, "id", "\x00\x00\x00\x17", 4, 0);

        block2 = enaml_new ();
        enaml_add_list (block2, "list", list, ENAML_FREE);
        enaml_add_block (block2, "block", block3, ENAML_FREE);
        enaml_add_string (block2, "name", "ack", 0);

        block_orig = enaml_new ();
        enaml_add_flag (block_orig, "warning");
        enaml_add_block (block_orig, "begin", block2, ENAML_FREE);

        block = enaml_copy_block (block_orig);
        enaml_free (block_orig);

	VERIFY (9, "block copying");

        enaml_free (block);
        return result;
}

#define COMPLEX_STRING "list:[9 \"house%21\" [3 2 1] %FF03] key int:500 string:\"satch\" blob: %00 nested-block: { inside: { deeper: { deeper: { useless } } } }"

static int
test10 (void)
{
        EnamlBlock *block;
        const char *expect = "blob:%00 int:500 list:[ 9 \"house!\" [ 3 2 1 ] %FF03 ] string:\"satch\" nested-block:{ inside:{ deeper:{ deeper:{ useless } } } } key ";
        int result;

        block = enaml_parse (COMPLEX_STRING);

        VERIFY (10, "simple parsing");

        enaml_free (block);
        return result;
}

static int
test11 (void)
{
        EnamlBlock *block;

        block = enaml_parse ("list:[1 2 3 4 5]");
        if (enaml_list_length (enaml_get_list (block, "list")) == 5) {
                PASS;
        } else {
                FAIL (11, "list size checking", "size != 5");
        }
}

static int
test12 (void)
{
        EnamlBlock *block;

        block = enaml_parse ("spaz:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{a:{");
        if (block == NULL) {
                PASS;
        } else {
                FAIL (12, "parser recursion", "parser didn't give up");
        }
}

static int
test13 (void)
{
        EnamlBlock *block;
        long num, zip;
        int blobsize;
        const char *blob;
        const char *string;

        block = enaml_parse ("num:5 other:\"94043\" name:\"smith\"");
        num = enaml_get_long (block, "num", 0);
        zip = enaml_get_long (block, "other", 0);
        blobsize = enaml_get_blob (block, "other", &blob);
        string = enaml_get_string (block, "name", NULL);

        if ((num == 5) && (zip == 94043) && (blobsize == 5) &&
            (strncmp (blob, "94043", 5) == 0) &&
            (strcmp (string, "smith") == 0)) {
                PASS;
        } else {
                FAIL (13, "simple fetching", "incorrect values");
        }
}

static int
test14 (void)
{
        EnamlBlock *block;

        block = enaml_parse ("list:[1 2 [3 4 5] 6 [7 [8 9]] ]");
        if ((enaml_get_long (block, "list.0", 0) != 1) ||
            (enaml_get_long (block, "list.3", 0) != 6)) {
                FAIL (14, "list fetching", "incorrect firstlevel values");
        }
        if ((enaml_get_long (block, "list.2.1", 0) != 4) ||
            (enaml_get_long (block, "list.4.0", 0) != 7) ||
            (enaml_get_long (block, "list.4.1.1", 0) != 9)) {
                FAIL (14, "list fetching", "incorrect depth values");
        }
        PASS;
}

static int
test15 (void)
{
        EnamlBlock *block, *block2;

        block = enaml_parse ("first:{ flag int:9 } second:{ inner:{ x:1 y:4 } flag }");
        block2 = enaml_get_block (block, "second.inner");

        if (block2 == NULL) {
                FAIL (15, "block fetching", "couldn't find inner block");
        }
        if (enaml_get_long (block2, "y", 0) != 4) {
                FAIL (15, "block fetching", "inner block is corrupted");
        }
        if (enaml_get_long (block, "second.inner.x", 0) != 1) {
                FAIL (15, "block fetching", "inner block member reference failed");
        }
        PASS;
}

static int
test16 (void)
{
        EnamlBlock *block;
        const char *expect = "test:{ other-flag int:615 flag string:\"hat\" } ";
        int result;

        block = enaml_buildf ("test", "flag other-flag:%f int:%d string:%s",
                              615, "hat");

        VERIFY (16, "simple buildf");

        enaml_free (block);
        return result;
}

static int
test17 (void)
{
        EnamlBlock *block, *block2;
        EnamlList *list1, *list2, *list3;
        const char *expect = "happy:{ const:{ thing:\"other\" } table:[ [ \"breakfast\" \"spuds n eggs\" ] [ \"lunch\" \"cuppa ramen\" ] [ \"dinner\" \"pizza\" ] ] coords:[ 19 11 ] inner-block:{ lamp:\"yes\" } } ";
        int result;

        list1 = enaml_new_list ();
        enaml_list_add_string (list1, "breakfast", 0);
        enaml_list_add_string (list1, "spuds n eggs", 0);
        list2 = enaml_new_list ();
        enaml_list_add_string (list2, "lunch", 0);
        enaml_list_add_string (list2, "cuppa ramen", 0);
        list3 = enaml_new_list ();
        enaml_list_add_string (list3, "dinner", 0);
        enaml_list_add_string (list3, "pizza", 0);
        block2 = enaml_new ();
        enaml_add_string (block2, "thing", "other", 0);
        block = enaml_buildf ("happy", "coords:%2d %s:%B const:%!B table:%*L", 19, 11,
                              "inner-block", enaml_buildf (NULL, "lamp:%s", "yes"),
                              block2, 3, list1, list2, list3);

        VERIFY (17, "complex buildf");

        enaml_free (block2);
        enaml_free (block);
        return result;
}


#define TIMED_STRING "list:[9 \"house%21\" %FF03] olist:[1 2 3] key int:500 string:\"satch\" blob: %00 nested-block: { inside: { deeper: { deeper: { useless } } } }"

static void
time_parse (void)
{
        struct timeval start, end;
        EnamlBlock *block;
        int i;
        long total_usec;
        float average_per_sec, average_usec;

        fprintf (stderr, "Timing parser ... ");
        fflush (stderr);

        gettimeofday (&start, NULL);
        for (i = 0; i < 10000; i++) {
                block = enaml_parse (TIMED_STRING);
                enaml_free (block);
        }
        gettimeofday (&end, NULL);

        total_usec = (end.tv_sec - start.tv_sec) * 1000000 + (end.tv_usec - start.tv_usec);
        average_per_sec = (10000.0 * 1000000.0) / (float)total_usec;
        average_usec = (float)total_usec / 10000.0;
        fprintf (stderr, " averaging %.2fusec/parse.\n", average_usec);
}

int
main (int argc, char **argv)
{
        fprintf (stderr, "Running enaml tests ");

        if (test1 () && test2 () && test3 () && test4 () && test5 () && test6 () &&
            test7 () && test8 () && test9 () && test10 () && test11 () && test12 () &&
            test13 () && test14 () && test15 () && test16 () && test17 ()) {
                fprintf (stderr, "\nTests passed!\n");
        }

        time_parse ();

        return 0;
}
